import numpy as np
import yaml
import itertools
import logging

from so3g.hk import load_range
from sotodlib import core

from scipy.interpolate import interp1d

logger = logging.getLogger(__name__)


def _get_swap_dict(d):
    """
    Swap dictionary keys and values; useful for organizing fields in
    config file by device name instead of alias in get_grouped_hkdata().
    """
    return {v: k for k, v in d.items()}


def _group_feeds(fields, alias_exists=False):
    """ Sort and group the HK feeds that are output from load_range()
    by device.

    Parameters:
        fields (dict) : dict of names of fields and/or aliases
        alias_exists (bool) : determines whether the field dict provided
            has aliases for the field names. If True, need to invert the
            dictionary to properly group by device. Grouping can only be done
            by knowing the exact field name. If False, no aliases were provided
            in the fields dictionary. Default is False.
    Returns:
        grouped_feeds (list): a sorted, grouped list of fields by device name;
            critical for eventually outputting an axismanager that's broken down
            per HK device

    """
    if alias_exists:
        sorted_fields = sorted(_get_swap_dict(fields))
    else:
        sorted_fields = sorted(fields)

    grouped_feeds = []
    key_func = lambda field: field.split('.')[0:4]
    for key, group in itertools.groupby(sorted_fields, key_func):
        grouped_feeds.append(list(group))

    return grouped_feeds


def _group_data(hkdata, field_dict=None, fields_only=False, alias_exists=False, config_exists=False):
    """ Sort hkdata that comes out of load_range depending
    on the parameters used for load_range()

    Parameters:
        hkdata (dict) : data output from load_range()
        field_dict (dict or None) : If None, this information already exists with
            the way hkdata was created; i.e., we already have a list of field names
            that we can immediately start grouping. If None, the argument for
            load_range() was only a list of field names (no aliases provided). If
            field_dict provided, aliases do exist for each field name, but
            are either coming from a config file input for load_range() or as a
            list of aliases input into load_range().
        fields_only (bool) : If True, only field names are provided with no aliases.
            Default is False.
        alias_exists (bool) : If True, aliases were given for each field name.
        config_exists (bool) : If True, load_range() was used using a config file;
            typically, config files have aliases so must follow a similar manipulation
            path as alias_exists. Default is False.

    Returns:
        get_grouped_data (list) : grouped list of data per HK feed/device;
        useful for turning into a grouped set of axismanagers

    """
    hknames = list(hkdata.keys())
    if fields_only:  # only field names provided, no aliases given for them
        grouped_feeds = _group_feeds(fields=hknames, alias_exists=False)

    if alias_exists or config_exists:
        # check for a field dict
        assert field_dict is not None

    if config_exists:
        online_fields = {}
        for i in range(len(hknames)):
            alias = hknames[i]
            field_name = field_dict[alias]
            field_info = {alias: field_name}
            online_fields.update(field_info)
    elif alias_exists:
        online_fields = _get_swap_dict(field_dict)

    if config_exists or alias_exists:
        fields_data = {}
        for alias in online_fields:
            time = hkdata[alias][0]
            data = hkdata[alias][1]
            field = online_fields[alias]

            info = {field: [alias, time, data]}
            fields_data.update(info)

        grouped_feeds = _group_feeds(fields=online_fields, alias_exists=True)

    grouped_data = []
    for group in grouped_feeds:
        device_data = {}
        for field in group:
            if fields_only:
                field_dict = {field: hkdata[field]}
                device_data.update(field_dict)
            if alias_exists or config_exists:
                field_dict = {field: fields_data[field]}
                device_data.update(field_dict)

        grouped_data.append(device_data)

    return grouped_data


def _check_hkdata(data):
    """
    A new bug was found in load_range() where if the same field name
    is given n times, load_range() returns an empty array of data n
    times. This is written to catch that bug and remove it from the
    hkdata dictonary until this bug is fixed.

    Parameters:
        data (dict) : dictionary of alias/fieldname, time, data from
            load_range()

    Returns:
        data (dict) : updated dictionary without the empty arrays

    """
    empty_keys = []
    for alias in list(data):
        if len(data[alias][0]) == 0:
            empty_keys.append(alias)

    for key in empty_keys:
        del data[key]


def sort_hkdata_fromconfig(start, stop, config):
    """
    Takes output from load_range(), reconfigures load_range() dictionary
    slightly. Groups HK field data by corresponding device/feed. Outputs
    that list of data.

    Parameters:
        start: Earliest time to search for data. Can be datetime objects,
            unix timestamps, int, floats.
        stop: Latest time to search for data. See start parameter above for
            note on time formats.
        config (str): Name of a .yaml file for loading fields and parsing
            information for housekeeping devices

    Returns:
        get_grouped_data (list): grouped list of data per HK feed/device;
            useful for turning into a grouped set of axismanagers, or for
            other grouped data analysis purposes.

    Notes
    -----

    An example config file looks like:

        data_dir: '/mnt/so1/data/chicago-latrt/hk/'

        field_list:
            'bf_4k' : 'observatory.LSA22HG.feeds.temperatures.Channel_06_T'
            'xy_stage_x': 'observatory.XYWing.feeds.positions.x'
            'xy_stage_y': 'observatory.XYWing.feeds.positions.y'

    """
    # call load_range()
    logger.debug("running load_range()")
    hkdata = load_range(start=start, stop=stop, config=config)
    _check_hkdata(hkdata)

    # load all fields from config file
    with open(config, 'r') as file:
        hkconfig = yaml.safe_load(file)

    field_dict = hkconfig['field_list']

    grouped_data = _group_data(hkdata=hkdata, field_dict=field_dict,
                              fields_only=False, alias_exists=False,
                              config_exists=True)

    return grouped_data


def sort_hkdata(start, stop, fields, data_dir, alias=None):
    """
    Takes output from load_range(), reconfigures load_range() dictionary
    slightly. Groups HK field data by corresponding device/feed. Outputs
    that list of data.

    Parameters:
        start: Earliest time to search for data. Can be datetime objects,
            unix timestamps, int, floats.
        stop: Latest time to search for data. See start parameter above for
            note on time formats.
        fields: (list) List of strings of Field names to query.
        alias (optional): (list) List of string of aliases of the field names

    Returns:
        get_grouped_data (list): grouped list of data per HK feed/device;
            useful for turning into a grouped set of axismanagers, or for
            other grouped data analysis purposes.

    """
    if alias is None:
        hkdata = load_range(start=start, stop=stop, fields=fields, data_dir=data_dir)
        _check_hkdata(hkdata)

        grouped_data = _group_data(hkdata=hkdata, fields_only=True,
                                   alias_exists=False, config_exists=False)

        return grouped_data

    elif alias is not None:
        hkdata = load_range(start=start, stop=stop, fields=fields,
                            alias=alias, data_dir=data_dir)

        field_dict = {}
        for i in range(len(fields)):
            assert len(fields) == len(alias)
            info = {fields[i]: alias[i]}
            field_dict.update(info)

        grouped_data = _group_data(hkdata=hkdata, field_dict=field_dict,
                                   fields_only=False, alias_exists=True,
                                   config_exists=False)

        return grouped_data


def make_hkaman(grouped_data, alias_exists=False, det_cosampled=False, det_aman=None):
    """
    Takes data from get_grouped_hkdata(), tests whether feed/device is
    cosampled, outputs axismanager(s) for either case.

    Parameters:
        grouped_data (list): grouped list of data per HK feed/device; output
            from get_grouped_hkdata()
        det_cosampled (bool): decides how to organize HK data into an AxisManager
            depending on user preference. Default set to False means HK data is
            output to an AxisManager per cosampled device, as well as 1 AxisManager
            per 'channel' for a non-cosampled device. If set to True, all HK data
            will be interpolated against detector timestreams.
        det_aman (NoneType): AxisManager of detector data from load_smurf().
            Only required if det_cosampled is True.

    Returns:
        merged_amans (AxisManager): AxisManager of HK AxisManagers. For HK
            cosampled data, it can contain 1 axismanager per cosmapled device,
            and 1 axismanger per 'channel' for a non-cosampled device.
            HK data that are cosampled to detector timestreams contain 1
            axismanager per HK device. All HK axismanagers are wrapped in 1 ultimate
            axismanager to streamline data analysis.
    """
    amans = []
    device_names = []
    for group in grouped_data:
        data = {}
        aliases = []
        times_hk = []
        times_det = {}
        for field in group:
            if alias_exists is True:
                # arrange data per device to input into aman
                alias = group[field][0]
                aliases.append(alias)

                time = group[field][1]
                times_hk.append(time)

                time_info = {alias: time}
                times_det.update(time_info)

                device_data = group[field][2]

                info = {alias: device_data}
                data.update(info)
            else:
                alias = field
                aliases.append(alias)

                time = group[field][0]
                times_hk.append(time)

                time_info = {alias: time}
                times_det.update(time_info)

                device_data = group[field][1]

                info = {alias: device_data}
                data.update(info)

        # only want HK data, checking for cases where HK data is cosampled
        if det_cosampled is False:
            try:
                # check for cosampled HK devices
                assert np.all([len(t) == len(times_hk[0]) for t in times_hk])
                # TODO: diff of times, if less than say 10%??, treat as cosampled
            except AssertionError:
                # if device not cosampled, make aman for each field in device/feed
                logger.debug("{} is not cosampled. Making separate axis managers \
                             for this case".format('.'.join(field.split('.')[0:3])))
                for field in group:
                    alias = group[field][0]
                    time = group[field][1]
                    data = group[field][2]

                    device_name = field.split('.')[1] + '_' + alias
                    device_names.append(device_name)

                    device_axis = 'hklabels_' + device_name
                    samps_axis = 'hksamps_' + device_name

                    hkaman = core.AxisManager(core.LabelAxis(alias, [alias]),
                                              core.OffsetAxis(samps_axis, len(time)))
                    hkaman.wrap('timestamps', time, [(0, samps_axis)])
                    hkaman.wrap(device_name, np.array([data]),
                                [(0, alias), (1, samps_axis)])
                    amans.append(hkaman)
            else:
                # if device cosampled, make 1 aman per device
                device_name = field.split('.')[1]
                device_names.append(device_name)
                device_axis = 'hklabels_' + device_name
                samps_axis = 'hksamps_' + device_name

                hkaman_cos = core.AxisManager(core.LabelAxis(device_axis, aliases),
                                              core.OffsetAxis(samps_axis, len(time)))
                hkaman_cos.wrap('timestamps', time, [(0, samps_axis)])
                hkaman_cos.wrap(device_name, np.array([data[alias] for alias in aliases]),
                                [(0, device_axis), (1, samps_axis)])
                amans.append(hkaman_cos)

        # want HK data cosampled to detector timestreams
        # not concerned with cosampeld HK devices here
        else:
            assert det_aman is not None, "Did not input a detector axis manager as an arg"

            device_name = field.split('.')[1]
            device_names.append(device_name)
            device_axis = 'hklabels_' + device_name

            data_interp = {}
            for alias in times_det:
                f_hkvals = interp1d(times_det[alias], data[alias], fill_value='extrapolate')
                hkdata_interp = f_hkvals(det_aman.timestamps)

                info = {alias: hkdata_interp}
                data_interp.update(info)

            hkaman = core.AxisManager(core.LabelAxis(device_axis, aliases),
                                      core.OffsetAxis('detector_timestamps',
                                                      len(det_aman.timestamps)))
            hkaman.wrap('timestamps', det_aman.timestamps, [(0, 'detector_timestamps')])
            hkaman.wrap(device_name, np.array([data_interp[alias] for alias in aliases]),
                        [(0, device_axis), (1, 'detector_timestamps')])
            amans.append(hkaman)

    # make an aman of amans
    merged_amans = core.AxisManager()
    [merged_amans.wrap(name, a) for (name, a) in zip(device_names, amans)]

    return merged_amans


def get_hkaman(start, stop, config=None, alias=None, fields=None, data_dir=None):
    """
    Wrapper to combine get_grouped_hkdata() and make_hkaman() to output one
    axismanager of HK axismanagers to streamline data analysis.

    Parameters:
        start: Earliest time to search for data. Can be datetime objects,
            unix timestamps, int, floats. Required for get_grouped_hkdata().
        stop: Latest time to search for data. See start parameter above for
            note on time formats. Required for get_grouped_hkdata()
        config (str): Filename of a .yaml file for loading fields and aliases.
            Required for get_grouped_hkdata()

    Notes
    -----

    An example config file looks like:

        data_dir: '/mnt/so1/data/chicago-latrt/hk/'

        field_list:
            'bf_4k' : 'observatory.LSA22HG.feeds.temperatures.Channel_06_T'
            'xy_stage_x': 'observatory.XYWing.feeds.positions.x'
            'xy_stage_y': 'observatory.XYWing.feeds.positions.y'
    """
    if config is not None:
        data = sort_hkdata_fromconfig(start=start, stop=stop, config=config)
        hkamans = make_hkaman(grouped_data=data, alias_exists=True,
                              det_cosampled=False)
        return hkamans

    elif fields is not None:
        # make sure you're working with the right components,
        # that you didn't provide any aliases
        if alias is None:
            data = sort_hkdata(start=start, stop=stop, fields=fields,
                               data_dir=data_dir)
            hkamans = make_hkaman(grouped_data=data, alias_exists=False,
                                  det_cosampled=False)
            return hkamans
        else:
            data = sort_hkdata(start=start, stop=stop, fields=fields,
                               data_dir=data_dir, alias=alias)
            hkamans = make_hkaman(grouped_data=data, alias_exists=True,
                                  det_cosampled=False)
            return hkamans


def get_detcosamp_hkaman(det_aman, config=None, alias=None, fields=None, data_dir=None):
    """
    Wrapper to combine get_grouped_hkdata() and make_hkaman() to output one
    axismanager of HK axismanagers that are cosampled to detector timestreams.

    Parameters:
        config (str): Filename of a .yaml file for loading HK fields and aliases.
            Required for get_grouped_hkdata().
        det_aman (AxisManager): detector data AxisManager from load_smurf().

    Notes
    -----

    An example config file looks like:

        data_dir: '/mnt/so1/data/chicago-latrt/hk/'

        field_list:
            'bf_4k' : 'observatory.LSA22HG.feeds.temperatures.Channel_06_T'
            'xy_stage_x': 'observatory.XYWing.feeds.positions.x'
            'xy_stage_y': 'observatory.XYWing.feeds.positions.y'
    """
    start = float(det_aman.timestamps[0])
    stop = float(det_aman.timestamps[-1])

    if config is not None:
        data = sort_hkdata_fromconfig(start=start, stop=stop, config=config)
        amans = make_hkaman(grouped_data=data, det_aman=det_aman,
                            alias_exists=True, det_cosampled=True)
        return amans
    elif fields is not None:
        if alias is None:
            data = sort_hkdata(start=start, stop=stop, fields=fields,
                               data_dir=data_dir)
            amans = make_hkaman(grouped_data=data, alias_exists=False,
                                det_cosampled=True, det_aman=det_aman)
            return amans
        else:
            data = sort_hkdata(start=start, stop=stop, fields=fields,
                               data_dir=data_dir, alias=alias)
            amans = make_hkaman(grouped_data=data, alias_exists=True,
                                det_cosampled=True, det_aman=det_aman)
            return amans



import numpy as np

def tag_substr(dest, tags, max_recursion=20):
    """ Do string substitution of all our tags into dest (in-place
    if dest is a dict). Used for context and data packaging tag replacements
    """
    assert(max_recursion > 0)  # Too deep this dictionary.
    if isinstance(dest, str):
        # Keep subbing until it doesn't change any more...
        new = dest.format(**tags)
        while dest != new:
            dest = new
            new = dest.format(**tags)
        return dest
    if isinstance(dest, list):
        return [tag_substr(x,tags) for x in dest]
    if isinstance(dest, tuple):
        return (tag_substr(x,tags) for x in dest)
    if isinstance(dest, dict):
        for k, v in dest.items():
            dest[k] = tag_substr(v,tags, max_recursion-1)
        return dest
    return dest

def get_coindices(v0, v1, check_unique=False):
    """Given vectors v0 and v1, each of which contains no duplicate
    values, determine the elements that are found in both vectors.
    Returns (vals, i0, i1), i.e. the vector of common elements and
    the vectors of indices into v0 and v1 where those elements are
    found.

    This routine will use np.intersect1d if it can.  The ordering of
    the results is different from intersect1d -- vals is not sorted,
    but rather the elements will appear in the same order that they
    were found in v0 (so that i0 is strictly increasing).

    The behavior is undefined if either v0 or v1 contain duplicates.
    Pass check_unique=True to assert that condition.

    """
    if check_unique:
        assert(len(set(v0)) == len(v0))
        assert(len(set(v1)) == len(v1))

    try:
        vals, i0, i1 = np.intersect1d(v0, v1, return_indices=True)
        order = np.argsort(i0)
        return vals[order], i0[order], i1[order]
    except TypeError:  # return_indices not implemented in numpy < 1.15
        pass

    # The old fashioned way
    v0 = np.asarray(v0)
    w0 = sorted([(j, i) for i, j in enumerate(v0)])
    w1 = sorted([(j, i) for i, j in enumerate(v1)])
    i0, i1 = 0, 0
    pairs = []
    while i0 < len(w0) and i1 < len(w1):
        if w0[i0][0] == w1[i1][0]:
            pairs.append((w0[i0][1], w1[i1][1]))
            i0 += 1
            i1 += 1
        elif w0[i0][0] < w1[i1][0]:
            i0 += 1
        else:
            i1 += 1
    if len(pairs) == 0:
        return (np.zeros(0, v0.dtype), np.zeros(0, int), np.zeros(0, int))
    pairs.sort()
    i0, i1 = np.transpose(pairs)
    return v0[i0], i0, i1


def get_multi_index(short_list, long_list):
    """For each item in long_list, determine the index at which it occurs
    in short_list.  Returns the equivalent of::

       np.array([short_list.index(x) if x in short_list else -1
                 for x in long_list])

    """
    w0 = sorted([(j, i) for i, j in enumerate(short_list)])
    w1 = sorted([(j, i) for i, j in enumerate(long_list)])
    i0, i1 = 0, 0
    indices = []
    while i0 < len(w0) and i1 < len(w1):
        if w0[i0][0] == w1[i1][0]:
            indices.append((w1[i1][1], w0[i0][1]))
            i1 += 1
        elif w0[i0][0] < w1[i1][0]:
            i0 += 1
        else:
            indices.append((w1[i1][1], -1))
            i1 += 1
    while i1 < len(w1):
        indices.append((w1[i1][1], -1))
        i1 += 1
    if len(indices) == 0:
        return np.zeros(0, int)
    indices.sort()
    return np.array([i0 for i1, i0 in indices])



import pytest
from unittest.mock import create_autospec
import tempfile
import yaml
import os
import datetime

from sotodlib.io.imprinter import *
from sotodlib.io.g3tsmurf_db import Files

from ._helpers import mpi_multi


# These tests are serial-only.
if mpi_multi():
    pytest.skip(allow_module_level=True)


@pytest.fixture(scope="session", autouse=True)
def imprinter():
    with tempfile.NamedTemporaryFile(mode='w') as f:
        g3tsmurf_file = tempfile.NamedTemporaryFile(mode='w', delete=False)
        g3tsmurf_config = {
            "data_prefix": "/data",
            "db_path": "test_g3tsmurf.db"
        }
        yaml.dump(g3tsmurf_config, g3tsmurf_file)
        g3tsmurf_file.close()
        
        im_config = {
            'db_path': '_pytest_imprinter.db',
            'g3tsmurf': g3tsmurf_file.name,
            'output_root': 'test_output',
            'tel_tubes': {
                'lat': {
                    'slots': ['slot1', 'slot2']
                },
            }
        }
        yaml.dump(im_config, f)
        f.flush()
        im = Imprinter(f.name, make_db=True)
    return im

# use this to clean up db file afterwards
@pytest.fixture(scope="session", autouse=True)
def delete_imprinter_db(imprinter):
    yield
    if os.path.exists(imprinter.db_path):
        os.remove(imprinter.db_path)
    if os.path.exists('test_g3tsmurf.db'):
        os.remove('test_g3tsmurf.db')

@pytest.fixture
def obsset():
    t0 = 1674090159
    t1 = 1674090259
    dt_t0 = datetime.datetime.fromtimestamp(t0, tz=datetime.timezone.utc)
    dt_t1 = datetime.datetime.fromtimestamp(t1, tz=datetime.timezone.utc)

    file = create_autospec(Files)
    file.start = dt_t0
    file.stop = dt_t1
    file.n_channels = 10
    
    obs1 = create_autospec(spec=G3tObservations, instance=True)
    obs1.obs_id = f"oper_slot1_{t0}" 
    obs1.files = [file] * 2
    obs1.timing = True
    obs2 = create_autospec(spec=G3tObservations, instance=True)
    obs2.obs_id = f"oper_slot2_{t0}"
    obs2.files = [file] * 2
    obs2.timing = True
    obsset = ObsSet([obs1, obs2], mode="oper", slots=["slot1", "slot2", "slot3"], tel_tube="lat")
    return obsset

def test_ObsSet(obsset):
    assert obsset.mode == "oper"
    assert obsset.slots == ["slot1", "slot2", "slot3"]
    assert obsset.tel_tube == "lat"
    assert obsset.obs_ids == ["oper_slot1_1674090159", "oper_slot2_1674090159"]
    assert obsset.get_id() == "oper_1674090159_lat_110"
    assert obsset.contains_stream("slot1") == True
    assert obsset.contains_stream("slot2") == True
    assert obsset.contains_stream("slot3") == False

def test_register_book(imprinter, obsset):
    book = imprinter.register_book(obsset)
    assert book.bid == "oper_1674090159_lat_110"
    assert book.status == UNBOUND
    assert book.obs[0].obs_id == "oper_slot1_1674090159"
    assert book.obs[1].obs_id == "oper_slot2_1674090159"
    assert book.tel_tube == "lat"
    assert book.type == "oper"
    expected_start = (
        datetime.datetime.fromtimestamp(1674090159, tz=datetime.timezone.utc)
        .replace(tzinfo=None)
    )
    expected_stop = (
        datetime.datetime.fromtimestamp(1674090259, tz=datetime.timezone.utc)
        .replace(tzinfo=None)
    )
    assert book.start == expected_start
    assert book.stop == expected_stop
    assert book.max_channels == 10
    assert book.message == ""
    assert book.slots == "slot1,slot2"
    assert imprinter.book_bound("oper_1674090159_lat_110") == False
    with pytest.raises(BookExistsError):
        imprinter.register_book(obsset)

def test_book_exists(imprinter):
    assert imprinter.book_exists("oper_1674090159_lat_110") == True
    assert imprinter.book_exists("oper_1674090159_lat_111") == False

def test_register_book_raises_BookExistsError(obsset, imprinter):
    # try to register the same ObsSet again, should raise BookExistsError
    with pytest.raises(BookExistsError):
        assert imprinter.register_book(obsset)

def test_get_session(imprinter):
    session = imprinter.get_session()
    assert session is not None

def test_get_book(imprinter):
    book = imprinter.get_book('oper_1674090159_lat_110')
    assert book is not None
    assert book.bid == 'oper_1674090159_lat_110'
    assert book.status == UNBOUND

def test_get_unbound_books(imprinter):
    # Create a new unbound book and add it to the database
    unbound_books = imprinter.get_unbound_books()
    # Assert that the book we just created is in the list of unbound books
    assert any([book.bid == 'oper_1674090159_lat_110' and book.status == UNBOUND for book in unbound_books])
    unbound_books[0].status = BOUND
    imprinter.session.commit()
    unbound_books = imprinter.get_unbound_books()
    # Assert that the book we just created is NOT in the list of unbound books
    assert all([book.bid != 'oper_1674090159_lat_110' or book.status != UNBOUND for book in unbound_books])

def test_book_bound(imprinter):
    books = imprinter.get_books()
    assert imprinter.book_bound(books[0].bid) == True
    books[0].status = UNBOUND
    imprinter.session.commit()
    assert imprinter.book_bound(books[0].bid) == False

def test_stream_timestamp():
    obs_id = 'oper_stream1_1674090159'
    stream, timestamp = stream_timestamp(obs_id)
    assert stream == 'stream1'
    assert timestamp == '1674090159'



from typing import Literal, Optional, Tuple, Union, cast, overload
from warnings import warn

import numpy as np
import scipy.ndimage as simg
import scipy.stats as ss
import so3g
from numpy.typing import NDArray
from scipy.sparse import csr_array
from skimage.restoration import denoise_tv_chambolle
from so3g.proj import Ranges, RangesMatrix
from sotodlib.core import AxisManager

from ..flag_utils import _merge


def std_est(
    x: NDArray[np.floating],
    ds: int = 1,
    win_size: int = 20,
    axis: int = -1,
    method: str = "median_unbiased",
) -> NDArray[np.floating]:
    """
    Estimate white noise standard deviation of data.
    More robust to jumps and 1/f then np.std()

    Arguments:

        x: Data to compute standard deviation of.

        ds: Downsample factor to use, does a naive slicing in blocks of ``win_size``.

        win_size: Window size to downsample by.

        axis: The axis to compute along.

        method: The method to pass to np.quantile.

    Returns:

        stdev: The estimated white noise standard deviation of x.
    """
    if ds > 2 * x.shape[axis]:
        ds = 1
    if ds > 1:
        x = np.moveaxis(x, axis, -1)
        x = x[..., : x.shape[-1] - (x.shape[-1] % win_size)]
        shape = list(x.shape) + [win_size]
        shape[-2] = -1
        x = x.reshape(tuple(shape))
        x = np.moveaxis(x, -2, 0)
        diff = np.diff(x[::ds], axis=-1)
        diff = np.moveaxis(diff, 0, -2)
        diff = diff.reshape(shape[:-1])
        diff = np.moveaxis(diff, -1, axis)
    else:
        diff = np.diff(x, axis=axis)
    # Find ~1 sigma limits of differenced data
    lims = np.quantile(
        diff,
        np.array([0.159, 0.841]),
        axis=axis,
        method=method,
    )
    # Convert to standard deviation
    return (lims[1] - lims[0]) / 8**0.5


def _jumpfinder(
    x: NDArray[np.floating],
    min_size: Optional[Union[float, NDArray[np.floating]]] = None,
    win_size: int = 20,
    exact: bool = False,
) -> NDArray[np.bool_]:
    """
    Matched filter jump finder.

    Arguments:

        x: Data to jumpfind on, expects 1D or 2D.

        min_size: The smallest jump size counted as a jump.

        win_size: Size of window used when peak finding.

        exact: Flag only the jump locations if True.
               If False flag the whole window (cheaper).

    Returns:

        jumps: Mask with the same shape as x that is True at jumps.
               Jumps within win_size of each other may not be distinguished.
    """
    if min_size is None:
        min_size = ss.iqr(x, -1)

    # Since this is intended for det data lets assume we either 1d or 2d data
    # and in the case of 2d data we find jumps along rows
    orig_shape = x.shape
    x = np.atleast_2d(x)
    dtype = x.dtype.name
    if len(x.shape) > 2:
        raise ValueError("x may not have more than 2 dimensions")
    if dtype == "float32":
        matched_filt = so3g.matched_jumps
    elif dtype == "float64":
        matched_filt = so3g.matched_jumps64
    else:
        raise TypeError("x must be float32 or float64")

    jumps = np.zeros(x.shape, dtype=bool, order="C")
    if x.shape[-1] < win_size:
        return jumps.reshape(orig_shape)

    msk = np.ptp(x, axis=-1) > min_size
    if not np.any(msk):
        return jumps.reshape(orig_shape)

    # Flag with a matched filter
    win_size += win_size % 2  # Odd win size adds a wierd phasing issue
    _x = np.asarray(x[msk], dtype=x.dtype, order="C")
    _jumps = np.empty_like(_x, dtype="int32", order="C")
    if isinstance(min_size, np.ndarray):
        _min_size = min_size[msk].astype(_x.dtype)
    elif min_size is None:
        raise TypeError("min_size is None")
    else:
        _min_size = (min_size * np.ones(len(_x))).astype(_x.dtype)
    matched_filt(_x, _jumps, _min_size, win_size)
    jumps[msk] = _jumps > 0

    if exact:
        structure = np.array([[0, 0, 0], [1, 1, 1], [0, 0, 0]])
        labels, _ = simg.label(jumps, structure)
        peak_idx = np.array(
            simg.maximum_position(
                np.diff(_x, axis=-1, prepend=np.zeros(len(_x))), labels
            )
        )
        jump_rows = [peak_idx[:, 0]]
        jump_cols = peak_idx[:, 1]
        jumps[:] = False
        jumps[np.flatnonzero(msk)[jump_rows], jump_cols] = True

    return jumps.reshape(orig_shape)


def jumpfix_subtract_heights(
    x: NDArray[np.floating],
    jumps: Union[Ranges, RangesMatrix, NDArray[np.bool_]],
    inplace: bool = False,
    heights: Optional[Union[NDArray[np.floating], csr_array]] = None,
    **kwargs,
) -> NDArray[np.floating]:
    """
    Naive jump fixing routine where we subtract known heights between jumps.
    Note that you should exepect a glitch at the jump locations.
    Works best if you buffer the jumps mask by a bit.

    Arguments:

        x: Data to jumpfix on, expects 1D or 2D.

        jumps: Boolean mask or Ranges(Matrix) of jump locations.
               Should be the same shape at x.

        inplace: Whether of not x should be fixed inplace.

        heights: Array of jump heights, can be sparse.
                 If None will be computed.

        **kwargs: Additional arguments to pass to estimate_heights if heights is None.

    Returns:

        x_fixed: x with jumps removed.
                 If inplace is True this is just a reference to x.
    """
    orig_shape = x.shape
    force_copy = not inplace
    if inplace and not x.flags['C_CONTIGUOUS']:
        warn("Requested in place jump fixing but signal is not contigious, a copy will be made as a buffer")
        force_copy = True
    x_use = np.asarray(np.atleast_2d(x), order="C")  # type: ignore
    x_fixed = x_use
    # When we switch to numpy 2.0 asarray can take copy=(False+force_copy if inplace else (True if force_copy else None))
    if force_copy and np.may_share_memory(x_use, x_fixed):
        x_fixed = x_fixed.copy()

    if isinstance(jumps, np.ndarray):
        jumps_rm = RangesMatrix.from_mask(np.atleast_2d(jumps))
    elif isinstance(jumps, Ranges):
        jumps_rm = RangesMatrix.from_mask(np.atleast_2d(jumps.mask()))
    elif isinstance(jumps, RangesMatrix):
        jumps_rm = jumps
    else:
        raise TypeError("jumps not RangesMatrix or convertable to RangesMatrix")

    if heights is None:
        heights = estimate_heights(x_fixed, jumps_rm.mask(), **kwargs)
    elif isinstance(heights, csr_array):
        heights = heights.toarray()
    heights = cast(NDArray[np.floating], heights)
    heights = np.asarray(heights, dtype=x.dtype, order="C")

    dtype = x.dtype.name
    if len(x.shape) > 2:
        raise ValueError("x may not have more than 2 dimensions")
    if dtype == "float32":
        fix = so3g.subtract_jump_heights
    elif dtype == "float64":
        fix = so3g.subtract_jump_heights64
    else:
        raise TypeError("x must be float32 or float64")
    fix(x_use, x_fixed, heights, jumps_rm)
    if force_copy and inplace:
        x[:] = x_fixed[:]

    return x_fixed.reshape(orig_shape)


def _make_step(signal: NDArray[np.floating], jumps: NDArray[np.bool_]):
    jumps = np.atleast_2d(jumps)
    jumps[:, [0, -1]] = False
    ranges = RangesMatrix.from_mask(jumps)
    signal_step = np.atleast_2d(signal.copy())
    samps = signal_step.shape[-1]
    for i, det in enumerate(ranges.ranges):
        for r in det.ranges():
            mid = int((r[0] + r[1]) / 2)
            signal_step[i, r[0] : mid] = signal_step[i, max(r[0] - 1, 0)]
            signal_step[i, mid : r[1]] = signal_step[i, min(r[1] + 1, samps - 1)]
    return signal_step.reshape(signal.shape)


def _diff_buffed(
    signal: NDArray[np.floating],
    jumps: Optional[NDArray[np.bool_]],
    win_size: int,
    make_step: bool,
) -> NDArray[np.floating]:
    win_size = int(win_size + win_size % 2)
    if jumps is not None and make_step:
        signal = _make_step(signal, jumps)
    diff_buffed = np.empty_like(signal)
    diff_buffed[..., :win_size] = 0
    diff_buffed[..., win_size:] = np.subtract(
        signal[..., win_size:],
        signal[..., : (-1 * win_size)],
        out=diff_buffed[..., win_size:],
    )

    return diff_buffed


def estimate_heights(
    signal: NDArray[np.floating],
    jumps: NDArray[np.bool_],
    win_size: int = 20,
    twopi: bool = False,
    make_step: bool = False,
    diff_buffed: Optional[NDArray[np.floating]] = None,
    clean: float = 80,
) -> NDArray[np.floating]:
    """
    Simple jump estimation routine.

    Arguments:

        signal: The signal with jumps.

        jumps: Boolean mask of jump locations in signal.

        win_size: Number of samples to buffer when estimating heights.

        twopi: If True, heights will be rounded to the nearest 2*pi

        make_step: If True jump ranges will be turned into clean step functions.

        diff_buffed: Difference between signal and a signal shifted by win_size.
                     If None will be computed.

        clean: Make each jump range a constant height set by this percentile.
               If this is 0 then it is ignored.

    Returns:

        heights: Array of jump heights.
    """
    if diff_buffed is None:
        diff_buffed = _diff_buffed(signal, jumps, win_size, make_step)

    jumps = np.atleast_2d(jumps)
    diff_buffed = np.atleast_2d(diff_buffed)
    if len(jumps.shape) > 2:
        raise ValueError("Only 1d and 2d arrays are supported")
    if twopi:
        diff_buffed = np.round(diff_buffed / (2 * np.pi)) * 2 * np.pi

    heights = np.zeros_like(jumps, dtype=float)
    heights[jumps] = diff_buffed[jumps]

    if clean > 0:
        ranges = RangesMatrix.from_mask(jumps)
        for i, det in enumerate(ranges.ranges):
            for r in det.ranges():
                sign = (
                    np.sign(np.mean(np.sign(heights[i, r[0] : r[1]]))) - 1
                ) / 2  # 0 pos, -1 neg
                heights[i, r[0] : r[1]] = np.percentile(
                    heights[i, r[0] : r[1]], abs(clean + 100 * sign)
                )

    return heights


def _filter(
    x: NDArray[np.floating],
    medfilt: int = 0,
    gaussian_width: float = 0,
    tv_weight: float = 0,
) -> NDArray[np.floating]:
    if gaussian_width > 0 or tv_weight > 0 or medfilt > 0:
        x = x.copy()
    if medfilt > 0:
        _size = medfilt - 1 + (medfilt % 2)
        size = np.ones(len(x.shape), dtype=int)
        size[-1] = _size
        x = simg.median_filter(x, size)
    if gaussian_width > 0:
        x = simg.gaussian_filter1d(x, gaussian_width, axis=-1, output=x).astype(float)
    if tv_weight > 0:
        channel_axis = 0
        if len(x.shape) == 1:
            channel_axis = None
        x = denoise_tv_chambolle(x, tv_weight, channel_axis=channel_axis)
    return x


@overload
def twopi_jumps(
    aman,
    signal,
    win_size,
    nsigma,
    atol,
    max_tol,
    fix: Literal[True],
    inplace,
    merge,
    overwrite,
    name,
    ds,
    **filter_pars,
) -> Tuple[RangesMatrix, csr_array, NDArray[np.floating]]: ...


@overload
def twopi_jumps(
    aman,
    signal,
    win_size,
    nsigma,
    atol,
    max_tol,
    fix: Literal[False] = False,
    inplace=...,
    merge=...,
    overwrite=...,
    name=...,
    ds=...,
    **filter_pars,
) -> Tuple[RangesMatrix, csr_array]: ...


def twopi_jumps(
    aman: AxisManager,
    signal: Optional[NDArray[np.floating]] = None,
    win_size: int = 20,
    nsigma: float = 5.0,
    atol: Optional[Union[float, NDArray[np.floating]]] = None,
    max_tol: float = 0.0314,
    fix: bool = True,
    inplace: bool = False,
    merge: bool = True,
    overwrite: bool = False,
    name: str = "jumps_2pi",
    ds: int = 10,
    **filter_pars,
) -> Union[
    Tuple[RangesMatrix, csr_array], Tuple[RangesMatrix, csr_array, NDArray[np.floating]]
]:
    """
    Find and optionally fix jumps that are height ~N*2pi.
    TOD is expected to have detectors with high trends already cut.
    Data is assumed to have units of phase here.

    Arguments:

        aman: The axis manager containing signal to find jumps on.

        signal: Signal to jumpfind on. If None than aman.signal is used.

        win_size: Size of window to use when looking for jumps.
                  This should be set to something of order the width of the jumps.

        nsigma: How many multiples of the white noise level to set to use to compute atol.
                Only used if atol is None.

        atol: How close the jump height needs to be to N*2pi to count.
              If set to None, then nsigma times the WN level of the TOD is used.
              Note that in general this is faster than nsigma.

        max_tol: Upper bound of the nsigma based thresh.
                 atol ignores this.

        fix: If True the jumps will be fixed by adding N*2*pi at the jump locations.

        inplace: If True jumps will be fixed inplace.

        merge: If True will wrap ranges matrix into ``aman.flags.<name>``

        overwrite: If True will overwrite existing content of ``aman.flags.<name>``

        name: String used to populate field in flagmanager if merge is True.

        ds: Downsample factor used when computing noise level, the actual factor used is `ds*win_size`.

        **filter_pars: Parameters to pass to _filter

    Returns:

        jumps: RangesMatrix containing jumps in signal,
               if signal is 1D Ranges in returned instead.
               Buffered to win_size.

        heights: csr_array of jump heights.

        fixed: signal with jump fixed. Only returned if fix is set.
    """
    if signal is None and isinstance(aman.signal, np.ndarray):
        signal = aman.signal
    if not isinstance(signal, np.ndarray):
        raise TypeError("Signal is not an array")
    if atol is None:
        atol = nsigma * std_est(
            signal.astype(float), ds=win_size * ds, win_size=win_size
        )
        np.clip(atol, 1e-8, max_tol)

    _signal = _filter(signal, **filter_pars)
    _signal = np.atleast_2d(_signal)
    if isinstance(atol, int) or isinstance(atol, float):
        atol = np.ones(len(_signal), float) * float(atol)
    elif np.isscalar(atol):
        raise TypeError(f"Invalid atol type: {type(atol)}")
    if len(atol) != len(signal):
        raise ValueError(f"Non-scalar atol provided with length {len(atol)}")

    _signal = np.asarray(_signal, dtype=_signal.dtype, order="C")
    heights = np.empty_like(_signal, order="C")
    atol = np.asarray(atol, dtype=_signal.dtype, order="C")
    if _signal.dtype.name == "float32":
        so3g.find_quantized_jumps(_signal, heights, atol, win_size, 2 * np.pi)
    elif _signal.dtype.name == "float64":
        so3g.find_quantized_jumps64(_signal, heights, atol, win_size, 2 * np.pi)
    else:
        raise TypeError("signal must be float32 or float64")

    # Shift things by half the window
    heights = np.roll(heights, -1 * int(win_size / 2), -1)
    heights[:, (-1 * int(win_size / 2)) :] = 0

    jumps = heights != 0
    jump_ranges = RangesMatrix.from_mask(jumps).buffer(int(win_size / 2))

    if merge:
        _merge(aman, jump_ranges, name, overwrite)

    if fix:
        fixed = jumpfix_subtract_heights(signal, jump_ranges, inplace, heights)
        return jump_ranges, csr_array(heights), fixed
    return jump_ranges, csr_array(heights)


@overload
def slow_jumps(
    aman,
    signal,
    win_size,
    thresh,
    abs_thresh,
    fix: Literal[True],
    inplace,
    merge,
    overwrite,
    name,
    clean,
    **filter_pars,
) -> Tuple[RangesMatrix, csr_array, NDArray[np.floating]]: ...


@overload
def slow_jumps(
    aman,
    signal,
    win_size,
    thresh,
    abs_thresh,
    fix: Literal[False] = False,
    inplace=...,
    merge=...,
    overwrite=...,
    name=...,
    clean=...,
    **filter_pars,
) -> Tuple[RangesMatrix, csr_array]: ...


def slow_jumps(
    aman: AxisManager,
    signal: Optional[NDArray[np.floating]] = None,
    win_size: int = 800,
    thresh: float = 20,
    abs_thresh: bool = True,
    fix: bool = True,
    inplace: bool = False,
    merge: bool = True,
    overwrite: bool = False,
    name: str = "jumps_slow",
    clean: float = 80,
    **filter_pars,
) -> Union[
    Tuple[RangesMatrix, csr_array], Tuple[RangesMatrix, csr_array, NDArray[np.floating]]
]:
    """
    Find and optionally fix slow jumps.
    This is useful for catching things that aren't really jumps but
    change the DC level in a jumplike way (ie: a short unlock period).

    Arguments:

        aman: The axis manager containing signal to find jumps on.

        signal: Signal to jumpfind on. If None than aman.signal is used.

        win_size: Size of window to use when looking for jumps.
                  This should be set to something of order the width of the jumps.

        thresh: ptp value at which to flag things as jumps.
                Default value is in phase units.
                You can also pass a percentile to use here if abs_thresh is False.

        abs_thresh: If True thresh is an absolute threshold.
                    If False it is a percential in range [0, 1).

        fix: If True the jumps will be fixed by adding N*2*pi at the jump locations.

        inplace: If True jumps will be fixed inplace.

        merge: If True will wrap ranges matrix into ``aman.flags.<name>``

        overwrite: If True will overwrite existing content of ``aman.flags.<name>``

        name: String used to populate field in flagmanager if merge is True.

        clean: Cleaning value to pass to ``estimate_heights``. See that function for details.

        **filter_pars: Parameters to pass to _filter

    Returns:

        jumps: RangesMatrix containing jumps in signal,
               if signal is 1D Ranges in returned instead.
               Buffered to win_size.

        heights: csr_array of jump heights.

        fixed: signal with jump fixed. Only returned if fix is set.
    """
    if signal is None and isinstance(aman.signal, np.ndarray):
        signal = aman.signal
    if not isinstance(signal, np.ndarray):
        raise TypeError("Signal is not an array")

    _signal = _filter(signal, **filter_pars)
    _signal = np.asarray(np.atleast_2d(_signal), dtype=_signal.dtype, order="C")

    # Block ptp
    dtype = _signal.dtype.name
    if len(_signal.shape) > 2:
        raise ValueError("signal may not have more than 2 dimensions")
    if dtype == "float32":
        get_ptp = so3g.block_minmax
    elif dtype == "float64":
        get_ptp = so3g.block_minmax64
    else:
        raise TypeError("signal must be float32 or float64")
    bptp = np.zeros_like(_signal, order="C")
    get_ptp(_signal, bptp, win_size, 2, 0)

    if not abs_thresh:
        thresh = float(np.quantile(bptp.ravel(), thresh))
    jumps = bptp > thresh

    jump_ranges = RangesMatrix.from_mask(jumps).buffer(int(win_size / 2))
    heights = estimate_heights(
        _signal, jump_ranges.mask(), win_size=win_size, make_step=True, clean=clean
    )

    if merge:
        _merge(aman, jump_ranges, name, overwrite)

    if fix:
        fixed = jumpfix_subtract_heights(signal, jump_ranges, inplace, heights)

        return jump_ranges, csr_array(heights), fixed
    return jump_ranges, csr_array(heights)


@overload
def find_jumps(
    aman,
    signal,
    min_sigma,
    min_size,
    win_size,
    exact,
    fix: Literal[False] = False,
    inplace=...,
    merge=...,
    overwrite=...,
    name=...,
    ds=...,
    clean=...,
    **filter_pars,
) -> Tuple[RangesMatrix, csr_array]: ...


@overload
def find_jumps(
    aman,
    signal,
    min_sigma,
    min_size,
    win_size,
    exact,
    fix: Literal[True],
    inplace,
    merge,
    overwrite,
    name,
    ds,
    clean,
    **filter_pars,
) -> Tuple[RangesMatrix, csr_array, NDArray[np.floating]]: ...


def find_jumps(
    aman: AxisManager,
    signal: Optional[NDArray[np.floating]] = None,
    min_sigma: Optional[float] = None,
    min_size: Optional[Union[float, NDArray[np.floating]]] = None,
    win_size: int = 20,
    exact: bool = False,
    fix: bool = False,
    inplace: bool = False,
    merge: bool = True,
    overwrite: bool = False,
    name: str = "jumps",
    ds: int = 10,
    clean: float = 80,
    **filter_pars,
) -> Union[
    Tuple[RangesMatrix, csr_array], Tuple[RangesMatrix, csr_array, NDArray[np.floating]]
]:
    """
    Find jumps in aman.signal_name with a matched filter for edge detection.
    Expects aman.signal_name to be 1D of 2D.

    Arguments:

        aman: axis manager.

        signal: Signal to jumpfind on. If None than aman.signal is used.

        min_sigma: Number of standard deviations to count as a jump, note that
                   the standard deviation here is computed by std_est and is
                   the white noise standard deviation, so it doesn't include
                   contributions from jumps or 1/f.
                   If min_size is provided it will be used instead of this.

        min_size: The smallest jump size counted as a jump.
                  By default this is set to None and min_sigma is used instead,
                  if set this will override min_sigma.
                  If both min_sigma and min_size are None then the IQR is used as min_size.

        win_size: Size of window used when peak finding.
                  Also used for height estimation, should be of order jump width.

        exact: If True search for the exact jump location.
               If False flag allow some undertainty within the window (cheaper).

        fix: Set to True to fix.

        inplace: Whether of not signal should be fixed inplace.

        merge: If True will wrap ranges matrix into ``aman.flags.<name>``

        overwrite: If True will overwrite existing content of ``aman.flags.<name>``

        name: String used to populate field in flagmanager if merge is True.

        ds: Downsample factor used when computing noise level, the actual factor used is `ds*win_size`.

        clean: Cleaning value to pass to ``estimate_heights``. See that function for details.

        **filter_pars: Parameters to pass to _filter

    Returns:

        jumps: RangesMatrix containing jumps in signal,
               if signal is 1D Ranges in returned instead.
               There is some uncertainty on order of a few samples.
               Jumps within a few samples of each other may not be distinguished.

        heights: csr_array of jump heights.

        fixed: signal with jump fixed. Only returned if fix is set.
    """
    if signal is None and isinstance(aman.signal, np.ndarray):
        signal = aman.signal
    if not isinstance(signal, np.ndarray):
        raise TypeError("Signal is not an array")

    orig_shape = signal.shape
    _signal = _filter(signal, **filter_pars)
    _signal = np.atleast_2d(_signal)

    if len(orig_shape) > 2:
        raise ValueError("Jumpfinder only works on 1D or 2D data")

    if min_size is None and min_sigma is not None:
        min_size = min_sigma * std_est(
            signal, ds=win_size * ds, win_size=win_size, axis=-1
        )
    if min_size is None:
        raise ValueError("min_size is somehow still None")
    if isinstance(min_size, np.ndarray) and np.ndim(min_size) > 1:  # type: ignore
        raise ValueError("min_size must be 1d or a scalar")
    elif isinstance(min_size, (float, int)):
        min_size = float(min_size) * np.ones(len(_signal))

    # Mean subtract, if we don't do this then when we cumsum we get floats
    # that are too big and lack the precicion to find jumps well
    _signal -= np.mean(_signal, axis=-1)[..., None]

    jumps = _jumpfinder(_signal, min_size, win_size, exact).reshape(orig_shape)

    jump_ranges = RangesMatrix.from_mask(jumps).buffer(int(win_size / 2))
    jumps = jump_ranges.mask()
    heights = estimate_heights(signal, jumps, win_size=win_size, clean=clean)

    if merge:
        _merge(aman, jump_ranges, name, overwrite)

    if fix:
        fixed = jumpfix_subtract_heights(
            signal, jump_ranges, inplace=inplace, heights=heights
        )
        return jump_ranges, csr_array(heights), fixed
    return jump_ranges, csr_array(heights)


def jumps_aman(
    aman: AxisManager, jump_ranges: RangesMatrix, heights: csr_array
) -> AxisManager:
    """
    Helper to wrap the jumpfinder outputs into a AxisManager for use with preprocess.


    Arguments:

        aman: AxisManager to steam axis information from.

        jump_ranges: RangesMatrix containing the jump flag.

        heights: csr_array of jump heights.

    Returns:

        jumps_aman: AxisManager containing the jump information wrapped in
                    'jump_flag' and 'jump_heights'.
    """
    jump_aman = AxisManager(aman.dets, aman.samps)
    jump_aman.wrap("jump_flag", jump_ranges, [(0, "dets"), (1, "samps")])
    jump_aman.wrap("jump_heights", heights, [(0, "dets"), (1, "samps")])

    return jump_aman



import argparse
import datetime as dt
from typing import Optional

from sotodlib.io.imprinter import Imprinter, BOUND, UPLOADED
from sotodlib.site_pipeline.util import init_logger

logger = init_logger(__name__, "update_librarian: ")

def main(config: str):
    """
    Update the book plan database with new data from the g3tsmurf database.

    Parameters
    ----------
    config : str
        Path to config file for imprinter
    """

    imprinter = Imprinter(
        config, 
        db_args={'connect_args': {'check_same_thread': False}},
    )

    session = imprinter.get_session()
    to_upload = imprinter.get_bound_books(session=session)

    failed_list = []
    for book in to_upload:
        success, err = imprinter.upload_book_to_librarian(
            book, session=session, raise_on_error=False
        )
        if not success:
            failed_list.append( (book.bid, err) )
        ## don't just continually fail
        if len(failed_list) > 5:
            break
    
    if len(failed_list) != 0:
        # raise the first error so we know something is wrong
        logger.error(f"Failed to upload books {[f[0] for f in failed_list]}")
        raise failed_list[0][1]




def get_parser(parser=None):
    if parser is None:
        parser = argparse.ArgumentParser()
    parser.add_argument(
        '--config', type=str, help="imprinter configuration file"
    )
    return parser


if __name__ == "__main__":
    parser = get_parser(parser=None)
    args = parser.parse_args()
    main(**vars(args))



from sotodlib import core
import numpy as np
import logging

logger = logging.getLogger(__name__)

# Note to future developers with a need for speed: there are two
# obvious places where OpenMP acceleration in a C++ routine would be
# helpful.  np.cov is not threaded, for some reason.  And in our mode
# removal we loop over detectors in python.  In both cases the
# algorithms are simple and parallelize trivially.  Thankfully
# np.linalg.eig is already threaded.


def get_pca_model(tod=None, pca=None, n_modes=None, signal=None,
                  wrap=None):
    """Convert a PCA decomposition into the signal basis, i.e. into
    time-dependent modes that one might use for cleaning or
    calibrating.

    The generalization of "common mode" computation is to convert the
    eigen-decomposition of the covariance matrix into a limited set of
    modes that explain a high fraction of the power in the input
    signals.

    Here we select the strongest eigenmodes from the PCA and compute
    the time-dependent modes and the projection of each detector
    signal onto each mode.  An approximation for the input signal may
    then be computed as

       signal_model = weights . modes

    where weights has shape (dets, eigen), modes has shape (eigen,
    samps), and . indicates matrix multipliaction.  The size of the
    eigen axis can be as large as the number of dets, but is typically
    smaller and in this routine is set by n_modes.

    Arguments:
        tod: AxisManager with dets and samps axes.
        pca: AxisManager holding a PCA decomposition, as returned by
            get_pca.  If not specified, it will be computed from the
            data.
        n_modes: integer specifying the number of modes to compute;
            the strongest modes are taken and this sets the size of
            the "eigen" axis in the output.  Defaults to len(dets),
            but beware that the resulting data object will be the same
            size as the input signal.
        signal: array of shape (dets, samps) that is used to construct
            the requested eigen modes.  If pca is not passed in, this
            signal is also used to compute the covariance for PCA.
        wrap: string; if specified then the returned result is also
            stored in tod under that name.

    Returns:
        An AxisManager with (dets, eigen, samps) axes.  The field
        'weights' has shape (dets, eigen) and the field 'modes' has
        shape (eigen, samps).

    """
    if pca is None:
        pca = get_pca(tod=tod, signal=signal)
    if n_modes is None:
        n_modes = pca.eigen.count

    mode_axis = core.IndexAxis('eigen', n_modes)
    output = core.AxisManager(tod.dets, mode_axis, tod.samps)
    if signal is None:
        signal = tod.signal

    R = pca.R[:, :n_modes]
    output.wrap('weights', R, [(0, 'dets'), (1, 'eigen')])
    output.wrap('modes', np.dot(R.transpose(), signal),
                [(0, 'eigen'), (1, 'samps')])
    if not(wrap is None):
        tod.wrap(wrap, output)
    return output


def get_pca(tod=None, cov=None, signal=None, wrap=None, mask=None):
    """Compute a PCA decomposition of the kind useful for signal analysis.
    A symmetric non-negative matrix cov of shape(n_dets, n_dets) can
    be decomposed into matrix R (same shape) and vector E (length
    n_dets) such that

        cov = R . diag(E) . R^T

    with . denoting matrix multiplication and T denoting matrix
    transposition.

    Arguments:
        tod: AxisManager with dets and samps axes.
        cov: covariance matrix to decompose; if None then cov is
            computed from tod.signal (or signal).
        signal: array of shape (dets, samps).  If cov is not provided,
            it will be computed from this matrix.  Defaults to
            tod.signal.
        wrap: string; if set then the returned result is also stored
            in tod under this name.
        mask: If specifed, a boolean array to select which dets are to
            be considered in the PCA decomp. This is achieved by
            modifying the cov to make the non-considered dets
            independent and low significance.

    Returns:
        AxisManager with axes 'dets' and 'eigen' (of the same length),
        containing fields 'R' of shape (dets, eigen) and 'E' of shape
        (eigen).  The eigenmodes are sorted from strongest to weakest.

    """
    if cov is None:
        # Compute it from signal
        if signal is None:
            signal = tod.signal
        cov = np.cov(signal)
        cov = np.atleast_2d(cov)  # In case there is exactly 1 det

    if mask is not None:
        var_min = min(np.diag(cov)[mask])
        for i in (~mask).nonzero()[0]:
            cov[i,:] = 0
            cov[:,i] = 0
            cov[i,i] = var_min * 1e-2

    dets = tod.dets

    mode_axis = core.IndexAxis('eigen', dets.count)
    output = core.AxisManager(dets, mode_axis)
    output.wrap('cov', cov, [(0, dets.name), (1, dets.name)])

    E, R = np.linalg.eigh(cov)
    E[np.isnan(E)] = 0.
    E, R = E.real, R.real

    idx = np.argsort(-E)

    output.wrap('E', E[idx], [(0, mode_axis.name)])
    output.wrap('R', R[:, idx], [(0, dets.name), (1, mode_axis.name)])
    if not(wrap is None):
        tod.wrap(wrap, output)
    return output


def add_model(tod, model, scale=1., signal=None, modes=None, weights=None):
    """Adds modeled modes, multiplied by some scale factor, into signal.

    Given a matrix of weights (dets, eigen) and modes (eigen, samps),
    this computes model signal:

        model_signal = weights . modes

    and adds the result to signal, scaled by scale:

        signal += scale * (weights . modes)

    The intended use is for PCA or common mode removal, for which user
    should pass scale = -1.

    Arguments:
        tod: AxisManager with dets and samps axes.
        model: AxisManager with shape (dets, eigen, dets) containing
            'modes' and 'weights' arrays.  This is the type of object
            returned by get_pca_model, for example.
        scale: Factor by which to scale model_signal before
            accumulating into signal.
        signal: array into which to accumulate model_signal; defaults
            to tod.signal.
        modes: array of modes with shape (eigen, samps), defaults to
            model.modes.
        weights: array of weights with shape (dets, eigen), defaults
            to model.weights.

    Returns:
        The signal array into which modes were accumulated.

    Notes:
        If you only want to operate on a subset of detectors, the most
        efficient solution is to pass in weights explicitly, forced to
        0 for the dets you want to omit.  There's an optimization to
        skip the computation for a detector if all the mode coupling
        weights are zero.

    """
    if signal is None:
        signal = tod.signal
    if modes is None:
        modes = model.modes
    if weights is None:
        weights = model.weights
    for i in range(model.dets.count):
        if np.all(weights[i] == 0):
            continue
        signal[i] += np.dot(weights[i, :] * scale, modes)
    return signal


def get_trends(tod, remove=False, size=1, signal=None):
    """Computes trends for each detector signal that remove the slope
    connecting first and last points, as well as the mean of the
    signal.  The returned object can be treated like PCA model (e.g.,
    it can be passed as the model input to add_model).

    Arguments:
        tod: AxisManager with dets and samps axes.
        remove: boolean, if True then the computed trends (and means)
            are removed from the signal.
        size: the number of samples on each end of the signal to use
            for trend level computation.  Defaults to 1.
        signal: array of shape (dets, samps) to compute trends on.
            Defaults to tod.signal.

    Returns:
        An AxisManager with (dets, eigen, samps) axes.  The field
        'weights' has shape (dets, eigen) and the field 'modes' has
        shape (eigen, samps).  There are two modes, which always have
        the same form: index 0 is all ones, and index1 is a smooth
        line from -0.5 to +0.5.

    """
    if signal is None:
        signal = tod.signal
    trends = core.AxisManager(tod.dets, core.IndexAxis('eigen', 2), tod.samps)
    modes = np.ones((trends.eigen.count, trends.samps.count))
    modes[1] = np.linspace(-0.5, 0.5, modes.shape[1])
    weights = np.empty((trends.dets.count, trends.eigen.count))
    weights[:, 0] = signal.mean(axis=1)
    size = max(1, min(size, signal.shape[1] // 2))
    weights[:, 1] = (signal[:, -size:].mean(axis=1) -
                     signal[:, :size].mean(axis=1))
    trends.wrap('modes', modes)
    trends.wrap('weights', weights)
    if remove:
        add_model(tod, trends, scale=-1, signal=signal)
    return trends


def pca_cuts_and_cal(tod, pca_aman, xfac=2, yfac=1.5, calc_good_medianw=False):
    """Finds the bounds of the pca box using IQR 
    statistics

    Parameters
    ----------
    tod : AxisManager
        observation axismanagers
    pca_aman : AxisManager
        output pca axismanager from get_pca_model
    xfac : int
        multiplicative factor for the width of the pca box.
        Default is 2.
    yfac : int
        multiplicative factor for the height of the box.
        Default is 1.5. 
    calc_good_medianw : bool
        If true, the resulting median weight is calculated 
        excluding bad dets. Default is false.

    Returns
    -------
    pca_relcal : AxisManager
        AxisManager pca and relcal information.
        
    """
    x = tod.det_cal.s_i
    y = np.abs(pca_aman.weights[:, 0])

    # remove positive Si values
    filt = np.where(x < 0)[0]
    xfilt = x[filt]
    yfilt = y[filt]

    # normalize weights
    ynorm = yfilt / np.median(yfilt)
    median_ynorm = np.median(ynorm)
    medianx = np.median(xfilt)

    # IQR of normalized weights
    iqry_norm = np.percentile(ynorm, 80) - np.percentile(ynorm, 20)

    # IQR of Si's
    iqrx = np.percentile(xfilt, 80) - np.percentile(xfilt, 20)

    # Find box heights using norm'd weights
    # Convert y bounds back to the scale of the raw weights
    ylb = (median_ynorm - yfac * iqry_norm) * np.median(yfilt)
    yub = (median_ynorm + yfac * iqry_norm) * np.median(yfilt)

    # Calculate box width
    xlb = medianx - xfac * iqrx
    xub = medianx + xfac * iqrx
    if xub > 0:
        mad = np.median(np.abs(xfilt - medianx))
        xub = medianx + xfac * mad

    xbounds = (xlb, xub)
    ybounds = (ylb, yub)

    # Get indices of the values in the box (indices are wrt `x` array)
    ranges = [x >= xlb,
              x <= xub,
              y >= ylb,
              y <= yub]
    m = ~(np.all(ranges, axis=0))

    if calc_good_medianw:
        medianw = np.median(pca_aman.weights[:,0][~m])
    else:
        medianw = np.median(pca_aman.weights[:,0])
    relcal_val = medianw/pca_aman.weights[:,0]

    pca_relcal = core.AxisManager(tod.dets, tod.samps)
    pca_relcal.wrap('pca_det_mask', m, [(0, 'dets')])
    pca_relcal.wrap('xbounds', np.array(xbounds))
    pca_relcal.wrap('ybounds', np.array(ybounds))
    pca_relcal.wrap('pca_mode0', pca_aman.modes[0], [(0, 'samps')])
    pca_relcal.wrap('pca_weight0', pca_aman.weights[:, 0], [(0, 'dets')])
    pca_relcal.wrap('relcal', relcal_val, [(0, 'dets')])
    pca_relcal.wrap('median', medianw)

    return pca_relcal


def get_common_mode(
    tod,
    signal='signal',
    method='median',
    wrap=None,
    weights=None,
):
    """Returns common mode timestream between detectors.
    This uses method 'median' or 'average' across detectors as opposed to a principle
    component analysis to get the common mode.

    Arguments
    ---------
        tod: axis manager
        signal: str, optional
            The name of the signal to estimate common mode or ndarray with shape of
            (n_dets x n_samps). Defaults to 'signal'.
        method: str
            method of common mode estimation. 'median' or 'average'.
        wrap: str or None.
            If not None, wrap the common mode into tod with this name.
        weights: array with dets axis
            If not None, estimate common mode by taking average with this weights.

    Returns
    -------
        common mode timestream

    """
    if isinstance(signal, str):
        signal = tod[signal]
    elif isinstance(signal, np.ndarray):
        if np.shape(signal) != (tod.dets.count, tod.samps.count):
            raise ValueError("When passing signal as ndarray shape must match (n_dets x n_samps).")
    else:
        raise TypeError("signal must be str, or ndarray")

    if method == 'median':
        if weights is not None:
            logger.warning('weights will be ignored because median method is chosen')
        common_mode = np.median(signal, axis=0)
    elif method == 'average':
        common_mode = np.average(signal, axis=0, weights=weights)
    else:
        raise ValueError("method flag must be median or average")
    if wrap is not None:
        tod.wrap(wrap, common_mode, [(0, 'samps')])

    return common_mode



"""Support for reading and writing simple metadata types to HDF5.

"Simple" metadata, at this point, means tabular data with columns that
mix extrinsic indices, intrinsic indices, and metadata fields.

String data is awkward in HDF5 / numpy / Python 3.  The approach
adopted here is to maintain data structures in unicode-compatible data
types (numpy 'U'), because this permits simple string comparisons
(such as mask = (results['band_name'] == 'rem')).  As a result, we
must convert 'U' fields to numpy 'S' type when writing to HDF5, and
then back to 'U' type on load from HDF5.  See
http://docs.h5py.org/en/stable/strings.html for a little more info.

"""

import numpy as np
import h5py

from sotodlib.core import AxisManager
from sotodlib.core.metadata import ResultSet, SuperLoader, LoaderInterface
import warnings

def write_dataset(data, filename, address, overwrite=False, mode='a'):
    """Write a metadata object to an HDF5 file as a single dataset.

    Args:
      data: The metadata object.  Currently only ResultSet and numpy
        structured arrays are supported.
      filename: The path to the HDF5 file, or an open h5py.File.
      address: The path within the HDF5 file at which to create the
        dataset.
      overwrite: If True, remove any existing group or dataset at the
        specified address.  If False, raise a RuntimeError if the
        write address is already occupied.
      mode: The mode specification used for opening the file
        (ignored if filename is an open file).

    """
    if isinstance(data, ResultSet):
        data = data.asarray(hdf_compat=True)
    elif isinstance(data, np.ndarray):
        pass
    else:
        raise TypeError("I do not know how to write type %s" % data.__class__)

    if isinstance(filename, str):
        context = h5py.File(filename, mode)
    else:
        # Wrap in a nullcontext so that the block below doesn't
        # close the File on exit.
        fout = filename
        filename = fout.filename
        context = _nullcontext(fout)

    with context as fout:
        if address in fout:
            if overwrite:
                del fout[address]
            else:
                raise RuntimeError(
                    f'Address {address} already exists in {filename}; '
                    f'pass overwrite=True to clobber it.')
        fout.create_dataset(address, data=data)


def read_dataset(fin, dataset):
    """Read a dataset from an HDF5 file and return it as a ResultSet.

    Args:
      fin: Filename or h5py.File open for reading.
      dataset: Dataset path.

    Returns:
      ResultSet populated from the dataset.  Note this is passed
      through _decode_array, so byte strings are converted to unicode.

    """
    if isinstance(fin, str):
        fin = h5py.File(fin, 'r')
    data = fin[dataset][()]
    data = _decode_array(data)
    rs = ResultSet(keys=list(data.dtype.names))
    for row in data:
        rs.rows.append(tuple(row))
    return rs


class DefaultHdfLoader(LoaderInterface):
    """Determine the type of H5 saved data and pass off the loading to the
    correct class.
    """
    def from_loadspec(self, load_params, **kwargs):
        with h5py.File(load_params['filename'], mode='r') as fin:
            # look for AxisManager save signature
            if '_axisman' in fin[ load_params['dataset'] ].attrs.keys():
                newload = AxisManagerHdfLoader()
                return newload.from_loadspec(load_params, **kwargs)
            else:
                newload = ResultSetHdfLoader()
                return newload.from_loadspec(load_params, **kwargs)

class AxisManagerHdfLoader(LoaderInterface):
    def from_loadspec(self, load_params, **kwargs):
        """ Generate an AxisManager from the load_params dictionary.
        """
        _kwargs = {k1: kwargs[k2] for k1, k2 in [('fields', 'load_fields')]
                   if k2 in kwargs}
        aman = AxisManager.load(load_params['filename'],
                                load_params['dataset'],
                                **_kwargs)
        return aman


class ResultSetHdfLoader(LoaderInterface):
    def _check_key_map(self, key_map, data_in):
        """Check that all keys in key_map are valid fields in data_in."""
        for key in key_map.keys():
            if key not in data_in.dtype.names:
                raise KeyError(f"{key} not included inside dataset with keys"
                               f"{data_in.dtype.names}")

    def _prefilter_data(self, data_in, load_fields=None):
        """When a dataset is loaded and converted to a structured numpy array,
        this function is called before the data are returned to the user. If
        'load_fields' has data for this instance, this function adds any fields
        specified within to the key map and checks that these fields are all
        valid fields in data_in.

        This function may be extended in subclasses, but you will
        likely want to call the super() handler before doing
        additional processing.
        """
        key_map = {}
        if load_fields is not None:
            for field in load_fields:
                if isinstance(field, dict):
                    key_map.update(field)
                else:
                    key_map[field] = field

            for k in data_in.dtype.names:
                if k not in key_map.keys():
                    key_map[k] = None

            self._check_key_map(key_map, data_in)

        return _decode_array(data_in, key_map=key_map)

    def _populate(self, data, keys=None, row_order=None):
        """Process the structured numpy array "data" and return a ResultSet.
        keys should be a list of field names to load from the data
        (default is None, which will load all fields).  row_order
        should be a list of indices into the desired rows of data
        (default is None, which will load all rows, in order).

        (This function can be overridden in subclasses, without
        calling the super.)

        """
        if keys is None:
            keys = [k for k in data.dtype.names]
        if row_order is None:
            row_order = range(len(data))
        rs = ResultSet(keys=keys)
        for i in row_order:
            rs.append({k: data[k][i] for k in rs.keys})
        return rs

    def from_loadspec(self, load_params, **kwargs):
        """Retrieve a metadata result from an HDF5 file.

        Arguments:
          load_params: an index dictionary (see below).

        Returns a ResultSet (or, for subclasses, whatever sort of
        thing is returned by self._populate).

        The "index dictionary", for the present case, may contain
        extrinsic and intrinsic selectors (for the 'obs' and 'dets'
        axes); it must also contain:

        - 'filename': full path to an HDF5 file.
        - 'dataset':  name of the dataset within the file.

        Note that this just calls batch_from_loadspec.

        """
        return self.batch_from_loadspec([load_params], **kwargs)[0]

    def batch_from_loadspec(self, load_params, **kwargs):
        """Retrieves a batch of metadata results.  load_params should be a
        list of valid index data specifications.  Returns a list of
        objects, corresponding to the elements of load_params.

        This function is relatively efficient in the case that many
        requests are made for data from a single file.

        """
        # Gather all relevant HDF5 files.
        file_map = {}
        for idx, load_par in enumerate(load_params):
            fn = load_par['filename']
            if fn not in file_map:
                file_map[fn] = []
            file_map[fn].append(idx)
        # Open each one and pull out the result.
        results = [None] * len(load_params)
        for filename, indices in file_map.items():
            with h5py.File(filename, mode='r') as fin:
                # Don't reread dataset unless it changes.
                last_dataset = None
                for idx in indices:
                    dataset = load_params[idx]['dataset']
                    if dataset is not last_dataset:
                        data = fin[dataset][()]
                        data = self._prefilter_data(data, **kwargs)
                        last_dataset = dataset

                    # Dereference the extrinsic axis request.  Every
                    # extrinsic axis key in the dataset must have a
                    # value specified in load_params.
                    ex_keys = []
                    mask = np.ones(len(data), bool)
                    for k in data.dtype.names:
                        if k.startswith('obs:'):
                            ex_keys.append(k)
                            mask *= (data[k] == load_params[idx][k])

                    # Has user made an intrinsic request as well?
                    for k in data.dtype.names:
                        if k.startswith('dets:') and k in load_params[idx]:
                            mask *= (data[k] == load_params[idx][k])

                    # TODO: handle non-concordant extrinsic /
                    # intrinsic requests.

                    # Output.
                    keys_out = [k for k in data.dtype.names
                                if k not in ex_keys]
                    results[idx] = self._populate(data, keys=keys_out,
                                                  row_order=mask.nonzero()[0])
        return results


def _decode_array(data_in, key_map={}):
    """Converts a structured numpy array to a structured numpy array,
    rewriting any 'S'-type string fields as 'U'-type string fields.

    Args:
      data_in: A structure numpy array (i.e. an ndarray with a dtype
        consisting of multiple named fields).
      key_map: A dict specifying how to rename fields.  Any key=>value
        pair here will cause data_in[key] to be written to
        data_out[value].  If value is None, the specified field will
        not be included in data_out.

    Returns:
        A new structured array, unless no changes are needed, in which
        case data_in is returned unmodified.

    """
    changes = False
    new_dtype = []
    columns = []
    for i, k in enumerate(data_in.dtype.names):
        key_out = key_map.get(k, k)
        changes = changes or (key_out != k)
        if key_out is None:
            continue
        if data_in.dtype[k].char == 'S':
            # Convert to unicode.
            columns.append(np.array([v.decode('ascii') for v in data_in[k]]))
            changes = True
        else:
            columns.append(data_in[k])
        if len(data_in[k].shape) == 1:
            new_dtype.append((key_out, columns[-1].dtype))
        else:
            new_dtype.append((key_out, columns[-1].dtype, data_in[k].shape[1:]))
    new_dtype = np.dtype(new_dtype)
    output = np.empty(data_in.shape, dtype=new_dtype)
    for k, c in zip(new_dtype.names, columns):
        output[k] = c
    return output


# Starting in Python 3.7, this can be had from contextlib.
class _nullcontext:
    def __init__(self, enter_result=None):
        self.enter_result = enter_result

    def __enter__(self):
        return self.enter_result

    def __exit__(self, *excinfo):
        pass


SuperLoader.register_metadata('DefaultHdf', DefaultHdfLoader)
SuperLoader.register_metadata('AxisManagerHdf', AxisManagerHdfLoader)
SuperLoader.register_metadata('ResultSetHdf', ResultSetHdfLoader)

# The old name... remove some day.
SuperLoader.register_metadata('PerDetectorHdf5', ResultSetHdfLoader)



"""Functions for working with an imprinter instance. These are generally
functions expected to be needed for humans making one-off changes to the
imprinter setup. Functions run as part of the automated pipeline are defined in imprinter.py
"""
from sqlalchemy import or_
import datetime as dt
import os
import os.path as op
import shutil
import time
import numpy as np

from sotodlib.io.imprinter import ( 
    Books,
    WONT_BIND,
    FAILED,
    UNBOUND,
    BOUND,
    UPLOADED,
    DONE,    
    ObsSet,
    G3tObservations,
)

from .load_smurf import (
    TimeCodes,
    SupRsyncType,
)

def set_book_wont_bind(imprint, book, message=None, session=None):
    """Change book status to WONT_BIND, meaning the files indicated into 
    this book will not be bound. .g3 files here will go into stray books.

    This function should not be integrated into automated data packaging.

    This book status is necessary because without it the automated data 
    packaging could continue to try and register and fail on the same book. 

    Parameters
    -----------
    imprint: Imprinter instance
    book: str or Book 
    message: str or None
        if not none, update book message to explain why the setting is changing
    session: BookDB session.
    
    """
    if session is None:
        session = imprint.get_session()

    if isinstance(book, str):
        book = imprint.get_book(book)

    if book.status != FAILED:
        imprint.logger.warning(
            f"Book {book} has not failed before being set not to WONT_BIND"
        )
    
    book_dir = imprint.get_book_abs_path(book)
    if op.exists(book_dir):
        print(f"Removing all files from {book_dir}")
        shutil.rmtree(book_dir)
    else: 
        print(f"Found no files in {book_dir} to remove")

    book.status = WONT_BIND 
    if message is not None:
        book.message = message
    session.commit()

def set_book_rebind(imprint, book, update_level2=False):
    """ Delete any existing staged files for a book as 
    set it's status to UNBOUND

    Parameters
    -----------
    imprint: Imprinter instance
    book: str or Book 
    update_level2: bool
        if true, update all the level 2 observation entries for the book. This is sometimes the reason books fail to bind
    """
    book_dir = imprint.get_book_abs_path(book)

    if op.exists(book_dir):
        print(f"Removing all files from {book_dir}")
        if os.path.isfile(book_dir):
            os.remove(book_dir)
        elif os.path.isdir(book_dir):
            shutil.rmtree(book_dir)
        else:
            print("How is this not a file or directory")
    else: 
        print(f"Found no files in {book_dir} to remove")

    book.status = UNBOUND
    imprint.get_session().commit()

    if update_level2:
        g3session, SMURF = imprint.get_g3tsmurf_session(return_archive=True)
        obs_dict = imprint.get_g3tsmurf_obs_for_book(book)
        print(f"Updating level 2 observations")
        for _, obs in obs_dict.items():
            SMURF.update_observation_files(obs, g3session, force=True)

def delete_level2_obs_and_book(imprint, book, session=None):
    """When there are certain readout communication errors, smurf will blindly 
    stream full rate data that is nonsense. For these, we don't want to keep 
    the data in stray books because it's nonsense and it's massive. Delete the 
    level 2 files, observation, and the book.
    """
    assert book.type == "oper", f"This function is for oper books"
    set_book_rebind(imprint, book)
    
    if session is None:
        session = imprint.get_session()
    print(f"Removing Level 2 for {book.bid}")
    obs_dict = imprint.get_g3tsmurf_obs_for_book(book)
    g3session, SMURF = imprint.get_g3tsmurf_session(
        return_archive=True
    )
    for _, obs in obs_dict.items():
        SMURF.delete_observation_files(
            obs, g3session, dry_run=False
        )  
    for o in book.obs:
        session.delete(o)
    session.delete(book)
    session.commit()   


def remove_level2_obs_from_book(imprint, book, bad_obs_id):
    """If one level2 observation is problematic, delete that book and re-register without it.
    """
    assert book.type == "obs", f"Book should be an 'obs' book"

    # keep staged clean
    set_book_rebind(imprint, book)
    odic = imprint.get_g3tsmurf_obs_for_book(book)
    obs_ids = [k for k in odic.keys()]
    assert bad_obs_id in obs_ids, f"{bad_obs_id} not in book {book.bid}"
    
    new_obs_list = [o for o in obs_ids if o != bad_obs_id]
    g3session = imprint.get_g3tsmurf_session()
    olist = [
        g3session.query(G3tObservations).filter(
            G3tObservations.obs_id == o
        ).one() for o in new_obs_list
    ]
    oset = ObsSet(
        olist,
        mode=book.type, 
        slots=imprint.tubes[book.tel_tube]['slots'],
        tel_tube=book.tel_tube
    )
    session = imprint.get_session()
    for o in book.obs:
        session.delete(o)
    session.delete(book)
    session.commit()
    return imprint.register_book(oset, session=session)

def find_overlaps(imprint, obs_id, min_ctime, max_ctime):
    """ helper function for when a level 2 observation could span multiple
    books. Creates a list of ObsSets with that obs_id, prints info to screen and
    returns the list. imprinter then has a function
    imprinter.register_book(obsset, commit=True) that can be used to register
    the desired observation. Example usage::

        rsets = utils.find_overlaps(
            imprint, 'obs_ufm_mv9_1714406208', <- obs id from error message
            min_ctime, max_ctime
        )
        imprint.register_book( rsets[0], commit=True)

    obs_id: level 2 obs_id that overlaps multiple observations
    """
    obsset = imprint.update_bookdb_from_g3tsmurf(
        min_ctime=min_ctime, max_ctime=max_ctime,
        return_obsset=True,
    )
    rsets = []
    for o in obsset:
        if obs_id in o.obs_ids:
            rsets.append(o)
    for i,r in enumerate(rsets):
        print(f"-----ObsSet {i}----------")
        for o in r:
            print("\t", o)

    return rsets

def block_fix_duplicate_timestamps(imprint):
    """Run through and fix all the books with duplicated ancillary timestamps"""

    failed_books = imprint.get_failed_books()
    fix_list = []
    for book in failed_books:
        if "duplicate timestamps" in book.message:
            fix_list.append(book)
    imprint.logger.info(
        f"Found {len(fix_list)} books with duplicate HK data to fix"
    )

    for book in fix_list:
        imprint.logger.info(f"Setting book {book.bid} for rebinding")
        set_book_rebind(imprint, book)
        imprint.logger.info(
            f"Binding book {book.bid} dropping duplicate HK data"
        )
        imprint.bind_book(book, ancil_drop_duplicates=True)

def block_set_rebind(imprint, update_level2=False):
    """Run through and set all books with files errors to be rebound"""

    failed_books = imprint.get_failed_books()
    
    fix_list = []
    for book in failed_books:
        if "Delete to retry bookbinding" in book.message:
            fix_list.append(book)
    imprint.logger.info(
        f"Found {len(fix_list)} books with files to be removed"
    )
    for book in fix_list:
        imprint.logger.info(f"Setting book {book.bid} for rebinding")
        set_book_rebind(imprint, book, update_level2=update_level2)    

def block_fix_bad_timing(imprint):
    """Run through and try rebinding all books with bad timing"""
    failed_books = imprint.get_failed_books()
    fix_list = []
    for book in failed_books:
        if "TimingSystemOff" in book.message:
            fix_list.append(book)
    for book in fix_list:
        imprint.logger.info(f"Setting book {book.bid} for rebinding")
        set_book_rebind(imprint, book)
        imprint.logger.info(
            f"Binding book {book.bid} while accepting bad timing"
        )
        imprint.bind_book(book, allow_bad_timing=True)

def report_smurf_timestamp_error(imprint, book):
    set_book_rebind(imprint, book)
    binder = imprint._get_binder_for_book(book)

    for sid, stream in binder.streams.items():
        try:
            stream.preprocess()
        except Exception as e:
            print(f"Did not finish preprocessing {sid} received error: {e}")

    binder.set_min_max_ctime()

    msg="          \t error ctime  \t missed time \t time since start \t time until stop\n"
    for sid, stream in binder.streams.items():
        msk = np.all([
            stream.times >= binder.min_ctime,
            stream.times <= binder.max_ctime,
        ], axis=0)
        idxs = np.where(np.diff(stream.times[msk]) <= 0)[0]
        msg += f"{sid}"
        if len(idxs) == 0:
            msg += "\tNo Errors"
        for x in idxs:
            msg += f"\t{stream.times[msk][x]:5f}\t{np.diff(stream.times[msk])[x]:5f}"
            msg += f"\t{stream.times[msk][x]-binder.min_ctime:5f}"
            msg += f"\t{binder.max_ctime-stream.times[msk][x]:5f}\n"
    
    ## reset to failed
    book.status = FAILED
    imprint.get_session().commit()
    
    del binder
    return msg

def report_acu_timestamp_error(imprint, book, max_dt=10):
    set_book_rebind(imprint, book)
    binder = imprint._get_binder_for_book(book)
    binder.preprocess()
    ancil = binder.ancil

    msg="          \t error ctime  \t missed time \t time since start \t time until stop\n"

    for fld in ['az', 'el', 'boresight', 'corotator_enc']:
        f = getattr(ancil.hkdata, fld)
        if f is None:
            continue
        msk = np.all([
            f.times >= binder.min_ctime,
            f.times <= binder.max_ctime,
        ], axis=0)
        msg += f"{f.addr}"
        if np.sum(msk) <= 2:
            msg += f"does not overlap detector data\n"
            continue
        idxs = np.where(np.diff(f.times[msk])>max_dt)[0]
        if len(idxs) == 0:
            msg += f"\t has no drops\n"
        for x in idxs:
            msg += f"\t{f.times[msk][x]:5f}\t{np.diff(f.times[msk])[x]:5f}"
            msg += f"\t{f.times[msk][x]-binder.min_ctime:5f}"
            msg += f"\t{binder.max_ctime-f.times[msk][x]:5f}\n"
    
    ## reset to failed
    book.status = FAILED
    imprint.get_session().commit()
    
    del binder
    return msg

def get_timecode_final(imprint, time_code, type='all'):
    """Check if all required entries in the g3tsmurf database are present for
    smurf or stray book regisitration.
    
    Parameters
    -----------
    imprint: Imprinter instance
    time_code: int
        5-digit ctime code to check for finalization
    type: str
        book type to check for
    
    Returns
    --------
    is_final, bool
    reason, int
        0 if the books are ready to be registered
        1 if we are missing metadata entries
        2 if we are missing file entries
        3 if unbound or failed books are preventing registration
    """
    assert type in ['all','stray','smurf']
    
    g3session, SMURF = imprint.get_g3tsmurf_session(return_archive=True)
    session = imprint.get_session()

    # this is another place I was reminded sqlite does not accept
    # numpy int32s or numpy int64s
    time_code = int(time_code)

    servers = SMURF.finalize["servers"]
    meta_agents = [s["smurf-suprsync"] for s in servers]
    files_agents = [s["timestream-suprsync"] for s in servers]

    tcm = g3session.query(TimeCodes.agent).filter(
        TimeCodes.timecode==time_code,
        TimeCodes.agent.in_(meta_agents),
        TimeCodes.suprsync_type == SupRsyncType.META.value,
    ).distinct().all()

    if type == 'smurf':
        if len(tcm) == len(meta_agents):
            return True, 0
        else:
            return False, 1
    
    if len(tcm) != len(meta_agents):
        return False, 1

    tcf = g3session.query(TimeCodes.agent).filter(
        TimeCodes.timecode == time_code,
        TimeCodes.agent.in_(files_agents),
        TimeCodes.suprsync_type == SupRsyncType.FILES.value,
    ).distinct().all()
    
    if len(tcf) != len(files_agents):
        return False, 2
    
    book_start = dt.datetime.utcfromtimestamp(time_code * 1e5)
    book_stop = dt.datetime.utcfromtimestamp((time_code + 1) * 1e5)

    q = session.query(Books).filter(
        Books.start >= book_start,
        Books.start < book_stop,
        or_(Books.type == 'obs', Books.type == 'oper'),
        or_(Books.status == UNBOUND, Books.status == FAILED), 
    )
    if q.count() > 0:
        return False, 3

    return True, 0

    
def set_timecode_final(imprint, time_code):
    """Add required entires to the g3tsmurf database in order to force the smurf
    and/or stray books to be created. Will be used if there are errors in
    suprsync data transfer.
    
    Parameters
    -----------
    imprint: Imprinter instance
    time_code: 5-digit ctime code to finalize
    """

    g3session, SMURF = imprint.get_g3tsmurf_session(return_archive=True)
    servers = SMURF.finalize["servers"]
    # this is another place I was reminded sqlite does not accept
    # numpy int32s or numpy int64s
    time_code = int(time_code)
    
    for server in servers:
        tcf = g3session.query(TimeCodes).filter(
            TimeCodes.timecode == time_code,
            TimeCodes.agent == server["timestream-suprsync"],
            TimeCodes.suprsync_type == SupRsyncType.FILES.value,
        ).first()
        if tcf is None:
            tcf = TimeCodes(
                stream_id="fake",
                suprsync_type=SupRsyncType.FILES.value,
                timecode=time_code,
                agent=server["timestream-suprsync"],
            )
            g3session.add(tcf)
            g3session.commit()

        tcm = g3session.query(TimeCodes).filter(
            TimeCodes.timecode==time_code,
            TimeCodes.agent == server["smurf-suprsync"],
            TimeCodes.suprsync_type == SupRsyncType.META.value,
        ).first()
        if tcm is None:
            tcm = TimeCodes(
                stream_id="fake",
                suprsync_type=SupRsyncType.META.value,
                timecode=time_code,
                agent=server["smurf-suprsync"],
            )
            g3session.add(tcm)     
            g3session.commit()


# Copyright (c) 2021 Simons Observatory.
# Full license can be found in the top level "LICENSE" file.

"""Check azss routines.

"""

import unittest
import numpy as np
from numpy.polynomial import legendre as L

from sotodlib import core
from sotodlib.tod_ops import azss


def get_scan(n_scans=33, scan_accel=0.025, scanrate=0.025,
             az0=2, az1=2.25, fs=200):
    """
    Returns timestamps and az positions for a CES scan.
    """
    t_scan = (az1-az0)/scanrate
    t_turnaround = 2*scanrate/scan_accel

    t_tot = np.asarray([])
    d_tot = np.asarray([])

    t_scan = np.arange(0, t_scan+1/fs, 1/fs)
    t_turn = np.arange(0, t_turnaround+1/fs, 1/fs)
    ds_right = az0 + t_scan*scanrate
    dt_right = az1 + t_turn * scanrate - 0.5 * scan_accel * t_turn ** 2
    ds_left = az1 - t_scan*scanrate
    dt_left = az0 - t_turn * scanrate + 0.5 * scan_accel * t_turn ** 2

    for i in range(n_scans):
        # right scan
        if i == 0:
            t_tot = np.concatenate((t_tot, t_scan))
        else:
            t_tot = np.concatenate((t_tot, t_tot[-1]+t_scan))
        d_tot = np.concatenate((d_tot, ds_right))
        # right turnaround
        t_tot = np.concatenate((t_tot, t_tot[-1]+t_turn))
        d_tot = np.concatenate((d_tot, dt_right))
        # left scan
        t_tot = np.concatenate((t_tot, t_tot[-1]+t_scan))
        d_tot = np.concatenate((d_tot, ds_left))
        # right turnaround
        t_tot = np.concatenate((t_tot, t_tot[-1]+t_turn))
        d_tot = np.concatenate((d_tot, dt_left))
    return t_tot, d_tot


def make_fake_azss_tod(max_mode=20, noise_amp=1, n_scans=10,
                      ndets=2, input_coeffs=None):
    """
    Makes an axis manager with azimuth synchronous signal
    in it, populated via legendre polynomials plus gaussian noise.
    """
    np.random.seed(0)
    ts, azpoint = get_scan(n_scans=n_scans)
    az_min, az_max = np.min(azpoint), np.max(azpoint)
    x = ( 2*azpoint - (az_min+az_max) ) / (az_max - az_min)

    fake_signal = np.zeros((ndets, len(ts)))
    if input_coeffs is None:
        input_coeffs = np.random.uniform(-10, 11, size=(ndets, max_mode+1))
    for nd in range(ndets):
        fake_signal[nd] += L.legval(x, input_coeffs[nd])
        noise = np.random.normal(0, noise_amp, size=len(ts))
        fake_signal[nd] += noise

    dets = ['det%i' % i for i in range(ndets)]
    mode_names = []
    for mode in range(max_mode+1):
        mode_names.append(f'legendre{mode}')

    tod_fake = core.AxisManager(core.LabelAxis('dets', vals=dets),
                                core.OffsetAxis('samps', count=len(ts)),
                                core.LabelAxis('modes', vals=mode_names))
    point = core.AxisManager(core.OffsetAxis('samps', count=len(ts)))
    tod_fake.wrap('boresight', point)
    tod_fake.boresight.wrap('az', azpoint, axis_map=[(0, 'samps')])
    tod_fake.wrap('timestamps', ts, axis_map=[(0, 'samps')])
    tod_fake.wrap('signal', np.atleast_2d(fake_signal),
                  axis_map=[(0, 'dets'), (1, 'samps')])
    tod_fake.wrap('input_coeffs', np.atleast_2d(input_coeffs),
                  axis_map=[(0, 'dets'), (1, 'modes')])
    return tod_fake


def get_coeff_metric(tod):
    """
    Evaluates fit is working by comparing coefficients in to out.
    """
    print(tod.input_coeffs[0])
    print(tod.azss_stats.coeffs[0])
    outmetric_num = (tod.azss_stats.coeffs - tod.input_coeffs)**2
    outmetric_denom = (tod.input_coeffs)**2
    return np.median(100*(outmetric_num/outmetric_denom))


class AzssTest(unittest.TestCase):
    "Test the Azimuth Synchronous Signal fitting functions"
    def test_fit(self):
        max_mode = 10
        tod = make_fake_azss_tod(noise_amp=0, n_scans=50, max_mode=max_mode)
        azss_stats, model_sig_tod = azss.get_azss(tod, method='fit', max_mode=max_mode, azrange=None, bins=100)
        ommedian = get_coeff_metric(tod)
        print(ommedian)
        self.assertTrue(ommedian < 5.0)

if __name__ == '__main__':
    unittest.main()



# This code has been ported over from enlib: 
# https://github.com/amaurea/enlib
# Ported over on 10/31/2023
# The original code is public domain. Modifications and additions 
# are subject to the main license in this repository.

"""This module provids a simple logging/output interface."""
import time
from logging import *
from pixell import memory, colors, mpi

class EnFilter(Filter):
    def __init__(self, rank=0):
        self.rank = rank
        try:
            # Try to get actual time since task start if possible
            import os, psutil
            p = psutil.Process(os.getpid())
            self.t0 = p.create_time()
        except ImportError:
            # Otherwise measure from creation of this filter
            self.t0 = time.time()
    def filter(self, record):
        record.rank  = self.rank
        record.wtime = time.time()-self.t0
        record.wmins = record.wtime/60.
        record.whours= record.wmins/60.
        record.mem   = memory.current()/1024.**3
        record.resmem= memory.resident()/1024.**3
        record.memmax= memory.max()/1024.**3
        return record

class QuietOthers(Filter):
    def __init__(self, rank=0, which=(INFO,)):
        self.rank  = rank
        self.which = which
    def filter(self, record):
        if self.rank == 0 or record.levelno not in self.which:
            return record

class ColoredFormatter(Formatter):
    def __init__(self, msg, colors={'DEBUG':colors.reset,'INFO':colors.lgreen,'WARNING':colors.lbrown,'ERROR':colors.lred, 'CRITICAL':colors.lpurple}):
        Formatter.__init__(self, msg)
        self.colors = colors
    def format(self, record):
        try:
            col = self.colors[record.levelname]
        except KeyError:
            col = colors.reset
        return col + Formatter.format(self, record) + colors.reset

default_format = "%(rank)3d %(wmins)7.2f %(resmem)5.2f %(mem)5.2f %(memmax)5.2f %(message)s"

def init(level=INFO, rank=mpi.COMM_WORLD.rank, file=None, fmt=default_format, color=True, shared=True):
    """Set up the root logger for output to console and file. Extra output records
    for mpi rank, time since process start and memory usage are added by default.
    Console output is colored by default, and info-level messages are muted from
    others than root for console output. File output is done independently for each
    task. If file-name can be a format string, which is then used as file%rank
    to produce the output file. Otherwise, a different output file must be passed
    for each mpi task. If file is None (the default), no file output is produced.
    The console level threshold is set by the level argument, which must be a
    python logging module level. The threshold does not apply to file output, where
    everything is output."""
    logger  = getLogger("enlib")
    logger.setLevel(DEBUG)
    if file:
        try:
            oname = file % rank
        except:
            oname = file
        fh = FileHandler(oname,mode="w")
        fh.setLevel(DEBUG)
        fh.addFilter(EnFilter(rank))
        formatter = Formatter(fmt)
        fh.setFormatter(formatter)
        logger.addHandler(fh)
    fclass = ColoredFormatter if color else Formatter
    ch = StreamHandler()
    ch.setLevel(level)
    ch.setFormatter(fclass(fmt))
    ch.addFilter(EnFilter(rank))
    if shared:
        ch.addFilter(QuietOthers(rank))
    logger.addHandler(ch)
    return logger

def verbosity2level(verbosity):
    if verbosity <= 0: return ERROR
    if verbosity <= 1: return INFO
    return DEBUG



# Copyright (c) 2023-2024 Simons Observatory.
# Full license can be found in the top level "LICENSE" file.
"""Timestream demodulation.
"""

import numpy as np
from astropy import units as u
import toast
import toast.ops
from toast.observation import default_values as defaults

from .. import ops as so_ops
from .job import workflow_timer
from .proc_noise_est import select_raw_noise_model


def setup_demodulate(operators):
    """Add commandline args and operators for demodulation.

    Args:
        operators (list):  The list of operators to extend.

    Returns:
        None

    """
    operators.append(
        toast.ops.Demodulate(
            name="demodulate",
            hwp_angle=defaults.hwp_angle,
            enabled=False,
        )
    )
    operators.append(
        toast.ops.NoiseEstim(
            name="demod_noise_estim",
            out_model="demod_noise",
            lagmax=1024,
            nbin_psd=64,
            nsum=1,
            naverage=64,
            enabled=False,
        )
    )
    # Fitting is only performed if demod_noise_estim is enabled
    operators.append(
        toast.ops.FitNoiseModel(
            name="demod_noise_estim_fit",
            out_model="demod_noise_fit",
            enabled=True,
        )
    )


@workflow_timer
def demodulate(job, otherargs, runargs, data):
    """Run timestream demodulation.

    Args:
        job (namespace):  The configured operators and templates for this job.
        otherargs (namespace):  Other commandline arguments.
        runargs (namespace):  Job related runtime parameters.
        data (Data):  The input data container.

    Returns:
        (Data):  The new, demodulated data.

    """
    log = toast.utils.Logger.get()
    timer = toast.timing.Timer()
    timer.start()

    # Configured operators for this job
    job_ops = job.operators

    if not job_ops.demodulate.enabled:
        # Short-circuit back
        return data

    # The pre-demodulation noise model to use
    noise_model = select_raw_noise_model(job, otherargs, runargs, data)

    # The Demodulation operator is special because it returns a
    # new TOAST data object
    job_ops.demodulate.stokes_weights = job_ops.weights_radec
    job_ops.demodulate.hwp_angle = job_ops.weights_radec.hwp_angle
    job_ops.demodulate.noise_model = noise_model
    new_data = job_ops.demodulate.apply(data)
    log.info_rank(
        "  Finished demodulation in",
        comm=data.comm.comm_world,
        timer=timer,
    )
    demod_weights = toast.ops.StokesWeightsDemod()
    job_ops.weights_radec = demod_weights
    if hasattr(job_ops, "binner"):
        job_ops.binner.stokes_weights = demod_weights
    if hasattr(job_ops, "binner_final"):
        job_ops.binner_final.stokes_weights = demod_weights

    if job_ops.demod_noise_estim.enabled:
        # Estimate the (mostly white) noise on the demodulated data
        log.info_rank(
            "  Building demodulated noise estimate...",
            comm=data.comm.comm_world,
        )
        job_ops.demod_noise_estim.apply(new_data)
        log.info_rank(
            "  Finished demodulated noise estimate in",
            comm=data.comm.comm_world,
            timer=timer,
        )

    # Only run noise fitting if estimation is enabled
    if job_ops.demod_noise_estim_fit.enabled:
        if not job_ops.demod_noise_estim.enabled:
            log.info_rank(
                "demod_noise_estim disabled, nothing to fit",
                comm=data.comm.comm_world,
            )
        else:
            # Create a fit to this noise model
            job_ops.demod_noise_estim_fit.noise_model = \
                job_ops.demod_noise_estim.out_model
            log.info_rank(
                "  Running fit to noise estimate...",
                comm=data.comm.comm_world,
            )
            job_ops.demod_noise_estim_fit.apply(new_data)
            log.info_rank(
                "  Fit 1/f noise model in",
                comm=data.comm.comm_world,
                timer=timer,
            )

    return new_data



# Copyright (c) 2023-2024 Simons Observatory.
# Full license can be found in the top level "LICENSE" file.
"""Simulated observing / scanning motion of the telescope.
"""

import numpy as np
from astropy import units as u
import toast
from toast.observation import default_values as defaults
import toast.ops

from .. import ops as so_ops
from ..instrument import simulated_telescope
from .job import workflow_timer


def setup_simulate_observing(parser, operators):
    """Add commandline args and operators for simulated observing.

    Args:
        parser (ArgumentParser):  The parser to update.
        operators (list):  The list of operators to extend.

    Returns:
        None

    """
    parser.add_argument(
        "--hardware", required=False, default=None, help="Input hardware file"
    )
    parser.add_argument(
        "--det_info_file",
        required=False,
        default=None,
        help="Input detector info file for real hardware maps",
    )
    parser.add_argument(
        "--det_info_version",
        required=False,
        default=None,
        help="Detector info file version such as 'Cv4'",
    )
    parser.add_argument(
        "--thinfp",
        required=False,
        type=int,
        help="Thin the focalplane by this factor",
    )
    parser.add_argument(
        "--bands",
        required=False,
        default=None,
        help="Comma-separated list of bands: LAT_f030 (27GHz), LAT_f040 (39GHz), "
        "LAT_f090 (93GHz), LAT_f150 (145GHz), "
        "LAT_f230 (225GHz), LAT_f290 (285GHz), "
        "SAT_f030 (27GHz), SAT_f040 (39GHz), "
        "SAT_f090 (93GHz), SAT_f150 (145GHz), "
        "SAT_f230 (225GHz), SAT_f290 (285GHz). ",
    )
    group = parser.add_mutually_exclusive_group(required=False)
    group.add_argument(
        "--telescope",
        default=None,
        help="Telescope to simulate: LAT, SAT1, SAT2, SAT3, SAT4.",
    )
    group.add_argument(
        "--tube_slots",
        default=None,
        help="Comma-separated list of optics tube slots: c1 (LAT_UHF), i5 (LAT_UHF), "
        " i6 (LAT_MF), i1 (LAT_MF), i3 (LAT_MF), i4 (LAT_MF), o6 (LAT_LF),"
        " ST1 (SAT_MF), ST2 (SAT_MF), ST3 (SAT_UHF), ST4 (SAT_LF).",
    )
    group.add_argument(
        "--wafer_slots", default=None, help="Comma-separated list of wafer slots. "
    )
    parser.add_argument(
        "--sample_rate",
        required=False,
        default=10,
        help="Sampling rate",
        type=float,
    )
    parser.add_argument(
        "--schedule", required=False, default=None, help="Input observing schedule"
    )
    parser.add_argument(
        "--sort_schedule",
        required=False,
        default=False,
        action="store_true",
        help="Sort the observing schedule by mean boresight RA.  "
        "This can limit the area of sky each process group deals with.",
    )
    parser.add_argument(
        "--realization",
        required=False,
        default=None,
        help="Realization index",
        type=int,
    )

    operators.append(
        toast.ops.SimGround(
            name="sim_ground",
            weather="atacama",
            detset_key="pixel",
            session_split_key="wafer_slot",
            enabled=False,
        )
    )
    operators.append(so_ops.CoRotator(name="corotate_lat"))
    operators.append(toast.ops.PerturbHWP(name="perturb_hwp", enabled=False))
    # Detector quaternion pointing
    operators.append(
        toast.ops.PointingDetectorSimple(
            name="det_pointing_azel_sim",
            boresight=defaults.boresight_azel,
            quats="quats_azel_sim",
            enabled=False,
        )
    )
    operators.append(
        toast.ops.PointingDetectorSimple(
            name="det_pointing_radec_sim",
            boresight=defaults.boresight_radec,
            quats="quats_radec_sim",
            enabled=False,
        )
    )


@workflow_timer
def simulate_observing(job, otherargs, runargs, comm):
    """Simulate the observing motion of the selected detectors with the schedule.

    Args:
        job (namespace):  The configured operators and templates for this job.
        otherargs (namespace):  Other commandline arguments.
        runargs (namespace):  Job related runtime parameters.
        comm (MPI.Comm):  The MPI world communicator (or None).

    Returns:
        (Data):  The simulated data.

    """
    log = toast.utils.Logger.get()
    timer = toast.timing.Timer()
    timer.start()

    # Configured operators for this job
    job_ops = job.operators

    # Make sure we have the required bands and schedule.  These might
    # not be set during a dry-run, but if we got this far they need to
    # be set.
    if otherargs.bands is None:
        msg = "bands must be specified for simulated observing"
        log.error_rank(msg, comm=comm)
        raise RuntimeError(msg)
    if otherargs.schedule is None:
        msg = "schedule must be specified for simulated observing"
        log.error_rank(msg, comm=comm)
        raise RuntimeError(msg)

    # Simulated telescope
    telescope = simulated_telescope(
        hwfile=otherargs.hardware,
        det_info_file=otherargs.det_info_file,
        det_info_version=otherargs.det_info_version,
        telescope_name=otherargs.telescope,
        sample_rate=otherargs.sample_rate * u.Hz,
        bands=otherargs.bands,
        wafer_slots=otherargs.wafer_slots,
        tube_slots=otherargs.tube_slots,
        thinfp=otherargs.thinfp,
        comm=comm,
    )
    ndet = len(telescope.focalplane.detectors)
    log.info_rank(
        f"  Simulated focalplane with {ndet} detectors "
        f"(thinfp = {otherargs.thinfp}) in",
        comm=comm,
        timer=timer,
    )

    # Load the schedule file
    schedule = toast.schedule.GroundSchedule()
    schedule.read(otherargs.schedule, comm=comm)
    log.info_rank("  Loaded schedule in", comm=comm, timer=timer)
    if otherargs.sort_schedule:
        schedule.sort_by_RA()
        log.info_rank("  Sorted schedule in", comm=comm, timer=timer)
    mem = toast.utils.memreport(msg="(whole node)", comm=comm, silent=True)
    log.info_rank(f"  After loading schedule:  {mem}", comm)

    # Get the process group size
    group_size = toast.job_group_size(
        comm,
        runargs,
        schedule=schedule,
        focalplane=telescope.focalplane,
        full_pointing=otherargs.full_pointing,
    )

    # Create the toast communicator.
    toast_comm = toast.Comm(world=comm, groupsize=group_size)

    # The data container
    data = toast.Data(comm=toast_comm)

    timer.clear()
    timer.start()

    job_ops.mem_count.prefix = "Before Simulation"
    job_ops.mem_count.apply(data)

    # Simulate the telescope pointing

    job_ops.sim_ground.telescope = telescope
    job_ops.sim_ground.enabled = True
    job_ops.sim_ground.schedule = schedule
    if job_ops.sim_ground.weather is None:
        job_ops.sim_ground.weather = telescope.site.name
    if otherargs.realization is not None:
        job_ops.sim_ground.realization = otherargs.realization
    log.info_rank("  Running simulated observing...", comm=comm)
    job_ops.sim_ground.apply(data)
    log.info_rank("  Simulated telescope pointing in", comm=comm, timer=timer)

    job_ops.mem_count.prefix = "After Scan Simulation"
    job_ops.mem_count.apply(data)

    # Apply LAT co-rotation
    if job_ops.corotate_lat.enabled:
        log.info_rank("  Running simulated LAT corotation...", comm=comm)
        job_ops.corotate_lat.apply(data)
        log.info_rank("  Apply LAT co-rotation in", comm=comm, timer=timer)

    # Perturb HWP spin
    if job_ops.perturb_hwp.enabled:
        log.info_rank("  Running simulated HWP perturbation...", comm=comm)
        job_ops.perturb_hwp.apply(data)
        log.info_rank("  Perturbed HWP rotation in", comm=comm, timer=timer)
    return data



""" Test the HWP angle solver
This tests part of site_pipeline.make_hwp_solutions.py
"""


import unittest
import numpy as np

from sotodlib import core
from sotodlib.hwp.g3thwp import G3tHWP


class HwpSolutionTest(unittest.TestCase):
    def test_empty_input(self):
        t = np.arange(0, 3600, 1/200)
        tod = core.AxisManager(core.OffsetAxis('samps', count=len(t)))
        tod.wrap('timestamps', t, [(0, 'samps')])

        solver = G3tHWP()  # default config
        solver.data = {}  # input empty data
        solution = solver.make_solution(tod)
        self.assertTrue(solution.logger_1 == 'No HWP data')
        self.assertTrue(solution.logger_2 == 'No HWP data')
        self.assertTrue(np.all(solution.hwp_angle == 0))

    def test_fake_input(self):
        period: int = 3600  # second
        bbbclk: float = 2e8  # Hz
        fhwp: float = 2.  # Hz
        num_edges: int = 1140  # number of edge per rotation

        t = np.arange(0, period, 1/200)
        tod = core.AxisManager(core.OffsetAxis('samps', count=len(t)))
        tod.wrap('timestamps', t, [(0, 'samps')])

        counter = np.cumsum(
            [bbbclk * 0.95 / num_edges / fhwp,
             bbbclk * 1.05 / num_edges / fhwp] * period * num_edges)
        index = np.arange(period * num_edges * fhwp)
        # reference slot
        counter = counter[(index % num_edges != 0) & (index % num_edges != 1)]
        time = counter / bbbclk * num_edges * fhwp
        slow_t = np.arange(period)

        data = {
            'counter_1': np.array([time, counter]),
            'counter_2': np.array([time, counter]),
            'counter_index_1': np.array([time, np.arange(len(counter))]),
            'counter_index_2': np.array([time, np.arange(len(counter))]),
            'irig_time_1': np.array([slow_t, slow_t]),
            'irig_time_2': np.array([slow_t, slow_t]),
            'rising_edge_count_1': np.array([slow_t, slow_t * bbbclk]),
            'rising_edge_count_2': np.array([slow_t, slow_t * bbbclk]),
            'quad_1': np.array([slow_t, [1] * period]),
            'quad_2': np.array([slow_t, [1] * period]),
            'approx_hwp_freq_1': np.array([slow_t, [fhwp] * period]),
            'approx_hwp_freq_2': np.array([slow_t, [fhwp] * period]),
            'pid_direction': np.array([slow_t, [1] * period]),
        }

        solver = G3tHWP()  # default config
        solver.data = data
        solution = solver.make_solution(tod)
        self.assertTrue(solution.logger_1 == 'Angle calculation succeeded')
        self.assertTrue(solution.logger_2 == 'Angle calculation succeeded')
        self.assertTrue(solution.version == 3)
        self.assertTrue(np.all(np.isfinite(solution.hwp_angle)))


if __name__ == '__main__':
    unittest.main()



import unittest
import os
import shutil
import tempfile

from sotodlib.core import metadata

from ._helpers import create_outdir, mpi_multi


@unittest.skipIf(mpi_multi(), "Running with multiple MPI processes")
class TestObsFileDB(unittest.TestCase):

    def setUp(self):
        self.test_filename = 'test_obsfiledb.sqlite'
        self.test_datatree = 'test_datatree'
        if os.path.exists(self.test_filename):
            os.remove(self.test_filename)
        self.test_dir = tempfile.mkdtemp()

    def tearDown(self):
        if os.path.exists(self.test_filename):
            os.remove(self.test_filename)
        if os.path.exists(self.test_dir):
            shutil.rmtree(self.test_dir)

    def get_simple_db(self, n_obs=2, n_detsets=2, n_dets=3, n_segs=3,
                      prepath=''):
        # Create database in RAM
        db = metadata.ObsFileDb()
        obs_ids = ['obs%i' % i for i in range(n_obs)]
        detsets = ['group%i' % i for i in range(n_detsets)]
        for di, d in enumerate(detsets):
            db.add_detset(d, ['det%i_%i' % (di, i) for i in range(n_dets)])
        for obs_id in obs_ids:
            for detset in detsets:
                for file_index in range(n_segs):
                    sample_index = file_index * 1000
                    filename = f'{prepath}{obs_id}_{detset}_{sample_index:04d}.g3'
                    db.add_obsfile(filename, obs_id, detset, sample_index)
        db.prefix = self.test_dir
        return db

    def test_000_basic(self):
        # Get example database
        n_obs, n_detsets = 4, 3
        db = self.get_simple_db(n_obs, n_detsets)
        for fmt in ['sqlite', 'dump', 'gz']:
            # Write database to disk.
            db.to_file(self.test_filename, fmt=fmt)

            # Load it up again.
            db2 = metadata.ObsFileDb.from_file(self.test_filename, fmt=fmt)

            # Check.
            assert (sorted(db.get_obs()) == sorted(db2.get_obs()))
            obs_id = db.get_obs()[0]
            assert (sorted(db.get_detsets(obs_id)) ==
                    sorted(db2.get_detsets(obs_id)))
            assert (db.get_detsets('not an obs') == [])

            # CLI
            if fmt == 'sqlite':
                metadata.obsfiledb.main([self.test_filename, 'files'])

    def test_010_remove(self):
        db = self.get_simple_db(4, 3)
        n0 = len(db.get_obs())
        obs_id = db.get_obs()[0]
        detsets = db.get_detsets(obs_id)

        # Drop obs_id.
        db.drop_obs(obs_id)
        n1 = len(db.get_obs())
        self.assertEqual(n0 - 1, n1)

        # Drop detset
        n0 = len(db.verify()['raw'])
        db.drop_detset(detsets[0])
        n1 = len(db.verify()['raw'])
        self.assertEqual(n1 * len(detsets),
                         n0 * (len(detsets) - 1))

    def test_020_update(self):
        n_detsets, n_segs = 4, 3
        db = self.get_simple_db(n_detsets=n_detsets, n_segs=n_segs)
        # Create all the missing files.
        for row in db.verify()['raw']:
            present, filename = row[:2]
            open(filename, 'w').close()
        # Rescan.
        results = db.verify()
        assert(all([r[0] for r in results['raw']]))
        # Remove a file and check that database can clean itself up.
        os.remove(results['raw'][0][1])
        results2 = db.verify()
        assert(results2['raw'][0][0] is False)
        db2 = db.copy()
        db2.drop_incomplete()

        self.assertEqual(db2.get_obs(), db.get_obs())
        self.assertEqual(len(db2.verify()['raw']),
                         len(db.verify()['raw']) - n_segs)

    def test_030_prefix(self):
        db = self.get_simple_db()
        # Create all the missing files.
        for row in db.verify()['raw']:
            present, filename = row[:2]
            open(filename, 'w').close()
        # Quick check we've done that properly...
        results = db.verify()
        assert(all([r[0] for r in results['raw']]))

        # Absolute paths should override db.prefix.
        db = self.get_simple_db(prepath=self.test_dir + '/')
        db.prefix = '/whatever'
        results = db.verify()
        assert(all([r[0] for r in results['raw']]))

    def test_040_lookups(self):
        db = self.get_simple_db(prepath='somewhere/else/')
        target = 'somewhere/else/obs0_group0_0000.g3'
        for item, result in enumerate([
                db.lookup_file(target, resolve_paths=False),
                db.lookup_file(self.test_dir + '/' + target),
                db.lookup_file(self.test_dir + '/' + target, resolve_paths=True),
                db.lookup_file('x/' + target, prefix='x/'),
                db.lookup_file('notx/../x/' + target, prefix='x/'),
                ]):
            self.assertEqual(result['obs_id'], 'obs0', 'Subcase #%i' % item)

        assert(db.lookup_file('obsXYZ_.g3', fail_ok=True) is None)
        with self.assertRaises(RuntimeError):
            db.lookup_file('obsXYZ_.g3')
        with self.assertRaises(RuntimeError):
            db.lookup_file('notx/../x/' + target, resolve_paths=False)

    def test_100_diff_patch(self):
        db0 = self.get_simple_db()
        db1 = self.get_simple_db()
        pd = metadata.obsfiledb.diff_obsfiledbs(db0, db1)
        assert not pd['different']

        db0 = self.get_simple_db(n_obs=2, n_detsets=2, n_dets=4)
        db1 = self.get_simple_db(n_obs=3, n_detsets=3, n_dets=5)

        pd = metadata.obsfiledb.diff_obsfiledbs(db1, db0)
        assert pd['different']
        assert not pd['patchable']

        pd = metadata.obsfiledb.diff_obsfiledbs(db0, db1)
        assert pd['different']
        assert pd['patchable']

        metadata.obsfiledb.patch_obsfiledb(pd['patch_data'], db0)
        pd = metadata.obsfiledb.diff_obsfiledbs(db0, db1)
        assert not pd['different']


if __name__ == '__main__':
    unittest.main()



""" Command-line interface for helping to clear out or clean up failed books. 

Make sure you have the DATAPKG_ENV environment variable set to fill in the tags present in the site-pipeline-configs config files

--action=failed: the script will cycle through failed books and
give the option to retry binding (useful if software changes mean we expect the
books to now be successfully bound), to skip binding those files (push
timestreams into the stray books), or do nothing.

--action=autofix: run through all the book errors and fix ones that have well
understood issues

"""

import os
import re
import argparse
import numpy as np
import datetime as dt
from typing import Optional

from sotodlib.io.imprinter import Imprinter, Books, FAILED
import sotodlib.io.imprinter_utils as utils

def fix_single_book(imprint:Imprinter, book:Books):
    print(
        f"Book ID: {book.bid}\n"
        f"Failed with message:\n"
        f"{book.message}"
    )
    resp = None
    while resp is None:
        resp = input(
            "Possible Actions to Take: "
            "\n\t1. Retry Binding"
            "\n\t2. Retry Binding with lvl2 updates"
            "\n\t3. Rebind with flag(s)"
            "\n\t4. Permanently Skip Binding"
            "\n\t5. Permanently Skip Binding. Delete lvl2."
            "\n\t6. Do Nothing"
            "\nInput Response: "
        )
        try: 
            resp = int(resp)
        except:
            print(f"Invalid Response {resp}")
            resp=None
        if resp is None or resp < 1 or resp > 6:
            print(f"Invalid Response {resp}")
            resp=None
    print(f"You selected {resp}")
    if resp == 1 or resp == 2:
        utils.set_book_rebind(imprint, book, update_level2=(resp==2) )
    elif resp == 3:
        utils.set_book_rebind(imprint, book)
        print("Set each flag [] represents defaults")
        ignore_tags = set_tag_and_validate( "Ignore Tags? (y/[n])" )
        ancil_drop_duplicates = set_tag_and_validate(
            "Drop Ancillary Duplicates? (y/[n])"
        )
        allow_bad_timing = set_tag_and_validate(
            "Allow Bad Timing? (Low Precision or Dropped Samples)? (y/[n])"
        )
        require_acu = set_tag_and_validate("Require ACU data? ([y]/n)")
        require_hwp = set_tag_and_validate("Require HWP data? ([y]/n)")
        require_monotonic_times = set_tag_and_validate(
            "Require Monotonic Housekeeping times? ([y]/n)"
        )
        imprint.bind_book(
            book, ignore_tags=ignore_tags, ancil_drop_duplicates=ancil_drop_duplicates,
            allow_bad_timing=allow_bad_timing,
            require_acu=require_acu,
            require_hwp=require_hwp,
            require_monotonic_times=require_monotonic_times,
        )
    elif resp == 4:
        utils.set_book_wont_bind(imprint, book)
    elif resp == 5:
        sure = set_tag_and_validate(
            "Are you sure you want to delete level 2?"
        )
        if sure:
            utils.delete_level2_obs_and_book(imprint, book)
        else:
            print("Skipping")
    elif resp == 6:
        pass
    else:
        raise ValueError("how did I get here?")

def set_tag_and_validate(msg):
    resp = input(msg)
    while (resp.lower() != 'n' and resp.lower() != 'y'):
        print(f"Response {resp} is invalid")
        resp = input(msg)
    return resp.lower() == 'y'

def check_failed_books(imprint:Imprinter):
    fail_list = imprint.get_failed_books()
    for book in fail_list:
        fix_single_book(imprint, book)

def _last_line(book):
    splits = book.message.split('\n')
    for s in splits[::-1]:
        if len(s) > 0:
            return s

class BookError:
    def __init__(self, imprint, book):
        self.imprint = imprint
        self.book = book
    
    @staticmethod
    def has_error(book):
        raise NotImplementedError
    def fix_book(self):
        raise NotImplementedError
    def report_error(self):
        raise NotImplementedError

class BookDirHasFiles(BookError):
    @staticmethod
    def has_error(book):
        return "BookDirHasFiles" in book.message
    def fix_book(self):
        utils.set_book_rebind(self.imprint, self.book)
        self.imprint.bind_book(self.book)
    def report_error(self):
        return f"{self.book.bid} has files already in staged" 

class SecondFail(BookError):
    @staticmethod
    def has_error(book):
        return "SECOND-FAIL" in book.message
    def fix_book(self):
        return 
    def report_error(self):
        msg = f"{self.book.bid} has already failed once during autofix\n\t" 
        msg += _last_line(self.book)
        return msg

class MissingReadoutIDs(BookError):
    @staticmethod
    def has_error(book):
        return 'MissingReadoutIDError' in book.message
    def fix_book(self):
        if self.book.type == "oper":
            utils.set_book_wont_bind(self.imprint, self.book)
            return
        remove_oid = re.findall(r"obs_ufm\S*_\S*", self.book.message)
        if len(remove_oid) == self.book.bid.split("_")[-1].count('1'):
            utils.set_book_wont_bind(self.imprint, self.book)
            return
        for oid in remove_oid:
                self.book = utils.remove_level2_obs_from_book(
                    self.imprint, self.book, oid
                )
    def report_error(self):
        return f"{self.book.bid} does not have readout ids"

class NoScanFrames(BookError):
    @staticmethod
    def has_error(book):
        return 'NoScanFrames' in book.message
    def fix_book(self):
        if (
            self.book.type == "oper" or 
            self.book.stop-self.book.start < dt.timedelta(minutes=5)
        ):
            utils.set_book_wont_bind(self.imprint, self.book)
        else :
            remove_oid = re.findall(r"obs_ufm\S*_\S*", self.book.message)
            if len(remove_oid) == self.book.bid.split("_")[-1].count('1'):
                utils.set_book_wont_bind(self.imprint, self.book)
                return
            for oid in remove_oid:
                self.book = utils.remove_level2_obs_from_book(
                    self.imprint, self.book,
                    oid
                )
    def report_error(self):
        return f"{self.book.bid} does not have detector data"

class NoHWPData(BookError):
    @staticmethod
    def has_error(book):
        return 'NoHWPData' in book.message
    def fix_book(self):
        utils.set_book_rebind(self.imprint, self.book)
        self.imprint.bind_book(self.book, require_hwp=False,)
    def report_error(self):
        return f"{self.book.bid} does not HWP data reading out"

class DuplicateAncillaryData(BookError):
    @staticmethod
    def has_error(book):
        return "DuplicateAncillaryData" in book.message
    def fix_book(self):
        utils.set_book_rebind(self.imprint, self.book)
        self.imprint.bind_book(self.book, ancil_drop_duplicates=True,)
    def report_error(self):
        return f"{self.book.bid} has duplicate ancillary data"

class NoMountData(BookError):
    @staticmethod
    def has_error(book):
        return "NoMountData" in book.message
    def fix_book(self):
        if self.book.type == 'obs':
            print("Cannot autofix obs books where the ACU was not reading out")
            return
        elif self.book.type == 'oper':
            utils.set_book_rebind(self.imprint, self.book)
            self.imprint.bind_book(self.book, require_acu=False,)
        else: 
            raise ValueError(f"What book got me here? {self.book.bid}")
    def report_error(self):
        return f"{self.book.bid} does not ACU data reading out"

class TimingSystemOff(BookError):
    """Two places this error is thrown. If we get to the timing counter
    incrementing error then we know the level 2 metadata isn't set correctly.
    This can mess up observation books because low precision timing means the
    different slots cannot be bound together.
    """
    @staticmethod
    def has_error(book):
        return 'TimingSystemOff' in book.message

    def fix_book(self):
        if "Timing counters not incrementing" in self.book.message:
            if self.book.type == 'obs':
                print(
                    "Cannot autofix obs books where timing counters aren't" 
                    " incrementing"
                )
            elif self.book.type == 'oper':
                utils.set_book_rebind(
                    self.imprint, self.book, update_level2=True
                )
                self.imprint.bind_book(self.book)
            else:
                raise ValueError(f"What book got me here? {self.book.bid}")
        else:
            utils.set_book_rebind(self.imprint, self.book)
            self.imprint.bind_book(self.book, allow_bad_timing=True,)
    
    def report_error(self):
        if "Timing counters not incrementing" in self.book.message:
            msg = f"{self.book.bid} has timing system errors not caught at level 2"
            if self.book.type == 'obs':
                msg += "\n\t LEVEL2-FAIL: probably have to delete book, update"
                msg += "update level 2, and replan books"
            return msg
        else:
            return f"{self.book.bid} has low precision timing"

class FileTooLargeError(BookError):
    @staticmethod
    def has_error(book):
        return "FileTooLargeError" in book.message
    def fix_book(self):
        utils.set_book_rebind(self.imprint, self.book)       
        utils.delete_level2_obs_and_book(self.imprint, self.book)
    def report_error(self):
        msg = f"{self.book.bid} has too large level 2 files\n"
        l = _last_line(self.book)
        msg += f"\t {l.split('/')[-1]}"
        return msg

class BadTimeSamples(BookError):
    max_drops_to_fix = 10000
    dropped = None

    @staticmethod
    def has_error(book):
        return "BadTimeSamples" in book.message

    def fix_book(self):
        if "time samples not increasing" in self.book.message:
            print("cannot autofix samples not increasing")
            return
        
        if self.dropped is None:
            self.report_error()
        
        ## if all our dropped values are less than the limit. Fix
        if np.all([x<=self.max_drops_to_fix for x in self.dropped.values()]):
            utils.set_book_rebind(self.imprint, self.book)        
            self.imprint.bind_book(self.book, allow_bad_timing=True,)
        ## if all our dropped values are more than the limit. Don't Bind
        elif np.all([x>self.max_drops_to_fix for x in self.dropped.values()]):
            print(f"All obs_ids have more than {self.max_drops_to_fix}"
                   " will not bind book")
            utils.set_book_wont_bind(self.imprint, self.book)
        ## if only some of the observations have dropped timing. remove them
        else:
            remove_oid = [k for k,x in self.dropped.items() 
                          if x>self.max_drops_to_fix]
            for oid in remove_oid:
                self.book = utils.remove_level2_obs_from_book(
                    self.imprint, self.book, oid
                )
                
    def report_error(self):
        msg = f"{self.book.bid} has dropped time samples\n"
        self.dropped = {}
        for l in self.book.message.split('\n'):
            if len(l)>0 and l[0] == '\t':
                self.dropped[l.split("\t")[1].split(":")[0]] = int(l.split("\t")[1].split(":")[-1])
                msg += l + "\n"
        return msg

AUTOFIX_ERRORS = [
    SecondFail,
    BookDirHasFiles,
    MissingReadoutIDs,
    NoScanFrames,
    NoHWPData,
    DuplicateAncillaryData,
    NoMountData,
    TimingSystemOff,
    FileTooLargeError,
    BadTimeSamples,
]

def process_book_failure(
    imprint:Imprinter, 
    book:Books, 
    report=True, 
    fix=False,
    error_list = None,
):
    if error_list is None:
        error_list = AUTOFIX_ERRORS

    found = False
    for error in error_list:
        if error.has_error(book):
            found = True
            err = error(imprint, book)
            if report:
                print( err.report_error() )
            if fix:
                try:
                    err.fix_book()  
                except:
                    print(f"Failed on {book.bid}. Recording Failure")
                    book.status = FAILED #just in case we cleared the error
                    book.message = book.message + \
                    f'\nSECOND-FAIL. Tried with to fix with {error.__name__}'
                    imprint.get_session().commit()
        if found:
            break ## stop on first found error
    return found
            

def autofix_failed_books(
    imprint:Imprinter, min_ctime=None, max_ctime=None,
    report=True, fix=True, error_list=None,
):
    session = imprint.get_session()
    failed = session.query(Books).filter(Books.status == FAILED)
    if min_ctime is not None:
        failed = failed.filter(
            Books.start >= dt.datetime.utcfromtimestamp(min_ctime),
        )
    if max_ctime is not None:
        failed = failed.filter(
            Books.start <= dt.datetime.utcfromtimestamp(max_ctime),
        )
    failed = failed.all()
    for book in failed:
        success = process_book_failure(
            imprint, book, 
            report=report, 
            fix=fix, 
            error_list=error_list
        )
        if not success:
            print(f"Cannot categorize {book.bid}")

def get_parser(parser=None):
    if parser is None:
        parser = argparse.ArgumentParser(
            description="Make sure you have the DATAPKG_ENV environment "
            "variable set to fill in the tags present in the "
            "site-pipeline-configs config files. This includes the 'configs' " 
            "tag pointing to your site-pipeline-configs directory"
        )
    parser.add_argument(
        "platform", 
        help="platform (lat, satp1, satp2, satp3)",
        type=str
    )
    parser.add_argument(
        "action", 
        help=" 'failed' or 'report' or 'autofix'",
        type=str
    )
    parser.add_argument(
        '--test-mode', action='store_true',
        help="if true, do not try to rebind books",
    )
    return parser

def main():

    parser = get_parser(parser=None)
    args = parser.parse_args()

    print("You must be running this from an account with write permissions to"
         " the database and the book write directory.")
    
    imprint = Imprinter.for_platform(args.platform)

    if args.action == 'failed':
        check_failed_books(imprint)
    elif args.action == 'report':
        autofix_failed_books(imprint, report=True, fix=False)
    elif args.action == 'autofix':
        autofix_failed_books(imprint, report=True, fix=True)
    else:
        raise ValueError(
            "Chosen action must be 'failed' 'autofix' or 'report'"
        )


"""
jobdb
=====

Connect to a database
---------------------

With sqlite::

  jdb = jobdb.JobManager(sqlite_file='jobsdb-test.db')

With postgres::

  db_url = sqy.engine.URL.create(
        "postgresql",
        username="abc",
        password="def",
        host="ghi",
        database="jkl")
  jdb = jobdb.JobManager(url=db_url)


Create new jobs
---------------

Check if job exists; create::

  jclass = 'my_analysis'
  tags = {'obs_id': '1234', 'wafer_slot': 10}
  if not len(jdb.get_jobs(jclass=jclass, tags=tags)):
    jdb.create_job(jclass, tags)

Identifying work
----------------

Find jobs to do::

  all_jobs = jdb.get_jobs(jclass=jclass, jstate='open')
  done_everything = len(all_jobs) == 0

Filter jobs to ones we might want to work on now::

  recent_memory = time.time() - 3600  # Don't retry until 1 hour has passed.
  to_do = [j for j in all_jobs \
           if j.jstate == 'open' and j.visit_time <= recent_memory]

Lock a job and work on it::

  with jdb.locked(to_do) as job:
    if job is not None:
      job.mark_visited()
      ok = do_a_job(job.tags)
      if ok:
        job.jstate = 'done'
      else:
        if job.visit_count > 5:
          job.state = 'failed'

Fixing things
-------------

Forcibly unlock all jobs (though feel free to be more targeted)::

  for j in jdb.get_jobs(jclass='my_analysis'):
    jdb.unlock(j.id)

Delete some jobs::

  for j in jdb.get_jobs(jclass='my_analysis'):
    jdb.remove_job(j.id)


"""

import sqlalchemy as sqy
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import sessionmaker, relationship, aliased
from contextlib import contextmanager
import argparse
import enum
import time
import os
import sys


__all__ = ['JobLockedError',
           'JobNotLockedError',
           'JobNotOwnedError',
           'JobNotUniqueError',
           'JState',
           'Job',
           'Tag',
           'JobManager']


class JobLockedError(Exception):
    pass


class JobNotLockedError(Exception):
    pass


class JobNotOwnedError(Exception):
    pass


class JobNotUniqueError(Exception):
    pass


class JState(enum.Enum):
    open = "open"
    done = "done"
    failed = "failed"
    ignored = "ignored"

    @classmethod
    def all(cls):
        return list(cls.__members__.keys())


Base = declarative_base()


class Job(Base):
    __tablename__ = 'jobs'
    id = sqy.Column(sqy.Integer, primary_key=True)
    jclass = sqy.Column(sqy.String)
    jstate = sqy.Column(sqy.Enum(JState))
    lock = sqy.Column(sqy.Float)
    lock_owner = sqy.Column(sqy.String)

    creation_time = sqy.Column(sqy.Float, default=0.)
    visit_time = sqy.Column(sqy.Float, default=0.)
    visit_count = sqy.Column(sqy.Integer, default=0)

    _tags = relationship("Tag", lazy='selectin', cascade='all')

    @property
    def tags(self):
        return Tag.items(self._tags)

    def __repr__(self):
        try:
            return (f'Job(id={self.id},jclass={self.jclass},'
                    f'jstate={self.jstate.value})')
        except Exception:
            return 'Job@%x' % id(self)

    def mark_visited(self, visit_count=None, now=None):
        if visit_count is None:
            visit_count = self.visit_count + 1
        if now is None:
            now = time.time()
        self.visit_count = visit_count
        self.visit_time = now


class Tag(Base):
    __tablename__ = 'tags'
    id = sqy.Column(sqy.Integer, primary_key=True)
    job_id = sqy.Column(sqy.Integer, sqy.ForeignKey('jobs.id'))
    key = sqy.Column(sqy.String)
    value = sqy.Column(sqy.String)

    def __repr__(self):
        return f'Tag(@job_id={self.id}:{self.key}={self.value})'

    def item(self):
        return self.key, self.value

    @staticmethod
    def items(tag_list):
        td = {}
        for t in tag_list:
            k, v = t.item()
            td[k] = v
        return td


class JobManager:
    def __init__(self, engine=None, url=None, sqlite_file=None):
        if engine is None:
            if url is None:
                if sqlite_file is None:
                    sqlite_file = 'my_jobs.db'
                engine = sqy.create_engine(
                    f'sqlite:///{sqlite_file}',
                    connect_args={'timeout': 3})
            else:
                engine = sqy.create_engine(url)
        self.engine = engine
        Base.metadata.create_all(self.engine)
        self.get_session = sessionmaker(bind=self.engine)

    def _lockstr(self):
        return 'me%i' % os.getpid()

    def create_job(self,
                   jclass=None,
                   tags={},
                   jstate=None,
                   creation_time=None,
                   visit_count=None,
                   visit_time=None):
        """Create a new job in the jobs table.

        Return the job.

        """
        existing = self.get_jobs(
            jclass, tags=tags,
            jstate=JState.all(), locked=None)
        if len(existing):
            raise JobNotUniqueError(
                'Found other records with same tags: '
                f'{[x.id for x in existing]}')

        if creation_time is None:
            creation_time = time.time()
        if jstate is None:
            jstate = JState.open
        tags = [Tag(key=k, value=str(v)) for k, v in tags.items()]
        job = Job(jclass=jclass,
                  jstate=jstate,
                  creation_time=creation_time,
                  visit_count=visit_count,
                  visit_time=visit_time,
                  _tags=tags)
        with self.session_scope() as session:
            session.add(job)
            session.commit()
            # Force a retrieval of job.id, before expunging --
            # otherwise some fields will dangle and sqlalchemy can
            # complain later about the detached object.
            job.id
            session.expunge(job)
        return job

    def get_jobs(self,
                 jclass=None,
                 tags=None,
                 jstate=None,
                 locked=None,
                 job_id=None):
        """Get a list of jobs meeting particular criteria.

        Note jclass and jstate can be string, a list (match any in
        list) or None (matches all job classes).

        The returned objects are detached from any database session,
        and should not be modified.  To operate on one or more of the
        jobs returned here, pass them first to the locked() context
        manager, which will check the records out from the database
        (if possible) for use in your process.

        """
        with self.get_session() as session:
            q = session.query(Job)
            if isinstance(jstate, str) and jstate == 'all':
                jstate = JState.all()
            if jclass is not None:
                if isinstance(jclass, str):
                    jclass = [jclass]
                q = q.filter(sqy.or_(*[Job.jclass == c for c in jclass]))
            if jstate is not None:
                if isinstance(jstate, str) or isinstance(jstate, JState):
                    jstate = [jstate]
                q = q.filter(sqy.or_(*[Job.jstate == s for s in jstate]))
            if tags is not None:
                for k, v in tags.items():
                    tags_alias = aliased(Tag)
                    q = q.join(tags_alias, Job.id == tags_alias.job_id).\
                        filter(sqy.and_(tags_alias.key == k,
                                        tags_alias.value == v))
            if job_id is not None:
                q = q.filter(Job.id.in_(job_id))
            if locked is True:
                q = q.filter(Job.lock != None)  # noqa: E711
            elif locked is False:
                q = q.filter(Job.lock == None)  # noqa: E711
            else:
                assert locked is None
            q = q.order_by(Job.jclass, Job.id)
            jobs = q.all()
            [session.expunge(j) for j in jobs]
        return jobs

    def lock(self, job_id, owner=None, force=False):
        """Lock a Job record by id.  If the Job is already locked, a
        JobLockedError is raised.

        Returns a Job object that has been expunged from the database
        session.  The object attributes can be modified, but won't be
        written back to the database unless the object is merged into
        a new session.

        """
        if owner is None:
            owner = self._lockstr()
        with self.session_scope() as session:
            q = session.query(Job)
            if force:
                q = q.filter(sqy.and_(Job.id == job_id))
            else:
                q = q.filter(sqy.and_(Job.id == job_id,
                                      Job.lock == None))  # noqa: E711
            n = q.update({Job.lock: time.time(), Job.lock_owner: owner})
            session.commit()

        with self.session_scope() as session:
            job = session.get(Job, job_id)
            session.expunge(job)

        if n == 0 or job.lock_owner != owner:
            raise JobLockedError()

        return job

    def unlock(self, job, merge=True):
        if not merge or isinstance(job, int):
            if isinstance(job, Job):
                job = job.id
            with self.session_scope() as session:
                session.query(Job).filter(Job.id == job).update(
                    {Job.lock: None, Job.lock_owner: None})
                session.commit()
        else:
            with self.session_scope() as session:
                j1 = session.query(Job).filter(Job.id == job.id).one()
                if j1.lock_owner is None:
                    raise JobNotLockedError()
                if j1.lock_owner != job.lock_owner:
                    raise JobNotOwnedError()
                job.lock = None
                job.lock_owner = None
                session.merge(job)
                session.commit()

    def clear_locks(self, jobs=None):
        if jobs is None:
            raise ValueError('Pass jobs="all" to clear all locks.')
        if jobs == 'all':
            jobs = [None]
        jobs = [(j.id if isinstance(j, Job) else j) for j in jobs]
        with self.session_scope() as s:
            for j in jobs:
                q = s.query(Job)
                if j is not None:
                    q = q.filter(Job.id == j)
                q.update({Job.lock: None, Job.lock_owner: None})

    def remove_job(self, job_id, check_locked=False):
        with self.session_scope() as session:
            if check_locked:
                q = session.query(Job).filter(
                    sqy.and_(Job.id == job_id,
                             Job.lock == None))  # noqa: E711
            else:
                q = session.query(Job).filter(Job.id == job_id)

            n = q.delete()
            session.commit()
        if n == 0:
            raise JobLockedError()

    @contextmanager
    def locked(self, jobs, count=None, owner=None):
        """Context Manager to grant exclusive access to one or more
        Job.  Job record is marked as locked, and this process may
        freely work on the job and alter the job data.  When execution
        leaves the context, the Job will be marked as unlocked.  Note
        the _database_ is only explicitly locked while this lock is
        being acquired and released.  In between, other entities can
        do other database stuff.

        Args:
          job (int, Job, or list): The Job to lock, or list of Jobs
            from which to try to draw lockable ones.
          count (int, None): The number of jobs to lock.  If specified
            as an integer, a list of up to that many jobs will be
            yielded.  If None, then a single job will be locked and
            yielded directly (if possible), otherwise None is yielded.
          owner (str): Override lock_owner string.

        Notes:
          If the job argument is a list, the function will try to
          yield one of the jobs from the list, skipping any that are
          locked by another session.  If no unlocked jobs are
          available, the usual exception will be raised or else a None
          yielded, as per none_if_locked argument.

        """
        if owner is None:
            owner = self._lockstr()
        if isinstance(jobs, (int, Job)):
            jobs = [jobs]
        locked = []
        for job in jobs:
            if len(locked) >= (1 if count is None else count):
                break
            if isinstance(job, Job):
                job = job.id
            try:
                j = self.lock(job)
            except JobLockedError:
                continue
            locked.append(j)
        try:
            if count is None:
                if len(locked):
                    yield locked[0]
                else:
                    yield None
            else:
                yield locked
        finally:
            for j in locked:
                self.unlock(j)

    def get_resource(self, jclass, n=None, jstate='open', tags={}):
        jobs = self.get_jobs(jclass, jstate=jstate, tags=tags)
        resources = []
        for job in jobs:
            if len(resources) >= (1 if n is None else n):
                break
            try:
                resources.append(ResourceHandle(self, job))
            except JobLockedError:
                pass
        if n is None:
            return resources[0] if len(resources) else None
        return resources

    @contextmanager
    def session_scope(self):
        session = self.get_session()
        try:
            yield session
            session.commit()
        except Exception:
            session.rollback()
            raise
        finally:
            session.close()


class ResourceHandle:
    job_id = None

    def __init__(self, db, job):
        self.db = db
        self.db.lock(job.id)
        self.job_id = job.id
        self.tags = dict(job.tags)

    def __del__(self):
        if self.job_id is not None:
            self.db.unlock(self.job_id)


# CLI support ...

def get_db(args):
    if args.sqlite_file:
        return JobManager(sqlite_file=args.sqlite_file)
    else:
        raise

def get_query_kwargs(args):
    kw = {}
    if args.jstate is not None:
        kw['jstate'] = args.jstate.split(',')
    if args.jclass is not None:
        kw['jclass'] = args.jclass.split(',')
    if args.job_id is not None:
        kw['job_id'] = []
        for _id in args.job_id:
            kw['job_id'].extend([_x.strip() for _x in _id.split(',')])
    return kw

def select(jdb, kw={}):
    jobs = jdb.get_jobs(**kw)
    for j in jobs:
        locked = f'L@{j.lock:.1f}' if j.lock else 'unlocked'
        visited = f'V%i@%.1f' % (j.visit_count, j.visit_time)
        print(f'{j.id:6} {j.jclass:20} {j.jstate.value:8} {locked} {visited} '
              f'{j.tags}')
    return jobs

def get_parser(parser=None):
    if parser is None:
        parser = argparse.ArgumentParser()

    parser.add_argument('-c', '--config-file')
    parser.add_argument('-d', '--sqlite-file')

    query_parser = argparse.ArgumentParser(add_help=False)
    query_parser.add_argument('--jclass')
    query_parser.add_argument('--jstate')
    query_parser.add_argument('--job-id', action='append', default=None)
    modep = parser.add_subparsers(
        dest='mode')

    p = modep.add_parser(
        'select',
        parents=[query_parser], help=
        "Print rows from the database."
        )
    p.add_argument('--action', default='none', choices=
                   ['none', 'list', 'clear-locks', 'delete',
                    'set-open', 'set-done', 'set-failed', 'set-ignored'])

    return parser

def cli(args=None):
    if args is None:
        args = sys.argv[1:]
    parser = get_parser()
    args = parser.parse_args(args=args)

    if args.mode == 'select':
        jdb = get_db(args)
        kw = get_query_kwargs(args)
        jobs = select(jdb, kw=kw)

        if args.action == 'none':
            print('Pass --action to do something to these records.')
        elif args.action == 'list':
            pass
        elif args.action == 'clear-locks':
            print('Clearing locks ...')
            jdb.clear_locks(jobs)
        elif args.action == 'delete':
            print('Removing jobs ...')
            for j in jobs:
                jdb.remove_job(j)
        elif args.action.startswith('set-'):
            for k in ['open', 'done', 'failed', 'ignored']:
                if args.action == f'set-{k}':
                    for j in jobs:
                        with jdb.locked(j) as _j:
                            if _j is None:
                                print(f'Failed to lock job_id={j.id}')
                            else:
                                print(f'Updating state for job_id={j.id}')
                                _j.jstate = k
        else:
            parser.error('Provide valid --action=... arg.')
    else:
        parser.error('Provide a mode')
